# -*- coding: utf-8 -*-

"""Utilities for ablation study configurations."""

import itertools as itt
import json
import logging
import pathlib
import time
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union
from uuid import uuid4

from ..training import _TRAINING_LOOP_SUFFIX
from ..utils import normalize_string

__all__ = [
    'ablation_pipeline',
    'ablation_pipeline_from_config',
    'prepare_ablation_from_config',
    'prepare_ablation_from_path',
    'prepare_ablation',
]

logger = logging.getLogger(__name__)

Mapping2D = Mapping[str, Mapping[str, Any]]
Mapping3D = Mapping[str, Mapping[str, Mapping[str, Any]]]


def ablation_pipeline(
    datasets: Union[str, List[str]],
    directory: Union[str, pathlib.Path],
    models: Union[str, List[str]],
    losses: Union[str, List[str]],
    optimizers: Union[str, List[str]],
    training_loops: Union[str, List[str]],
    *,
    epochs: Optional[int] = None,
    create_inverse_triples: Union[bool, List[bool]] = False,
    regularizers: Union[None, str, List[str]] = None,
    negative_sampler: Union[str, None] = None,
    evaluator: Optional[str] = None,
    stopper: Optional[str] = 'NopStopper',
    model_to_model_kwargs: Optional[Mapping2D] = None,
    model_to_model_kwargs_ranges: Optional[Mapping2D] = None,
    model_to_loss_to_loss_kwargs: Optional[Mapping3D] = None,
    model_to_loss_to_loss_kwargs_ranges: Optional[Mapping3D] = None,
    model_to_optimizer_to_optimizer_kwargs: Optional[Mapping3D] = None,
    model_to_optimizer_to_optimizer_kwargs_ranges: Optional[Mapping3D] = None,
    model_to_negative_sampler_to_negative_sampler_kwargs: Optional[Mapping3D] = None,
    model_to_negative_sampler_to_negative_sampler_kwargs_ranges: Optional[Mapping3D] = None,
    model_to_training_loop_to_training_loop_kwargs: Optional[Mapping3D] = None,
    model_to_training_loop_to_training_kwargs: Optional[Mapping3D] = None,
    model_to_training_loop_to_training_kwargs_ranges: Optional[Mapping3D] = None,
    model_to_regularizer_to_regularizer_kwargs: Optional[Mapping3D] = None,
    model_to_regularizer_to_regularizer_kwargs_ranges: Optional[Mapping3D] = None,
    evaluator_kwargs: Optional[Mapping[str, Any]] = None,
    evaluation_kwargs: Optional[Mapping[str, Any]] = None,
    stopper_kwargs: Optional[Mapping[str, Any]] = None,
    n_trials: Optional[int] = 5,
    timeout: Optional[int] = 3600,
    metric: Optional[str] = 'hits@10',
    direction: Optional[str] = 'maximize',
    sampler: Optional[str] = 'random',
    pruner: Optional[str] = 'nop',
    metadata: Optional[Mapping] = None,
    save_artifacts: bool = True,
    move_to_cpu: bool = True,
    dry_run: bool = False,
    best_replicates: Optional[int] = None,
    discard_replicates: bool = False,
    create_unique_subdir: bool = False,
):
    """Run ablation study.

    :param datasets: A dataset name or list of dataset names.
    :param directory: The directory in which the experimental artifacts will be saved.
    :param models: A model name or list of model names.
    :param losses: A loss function name or list of loss function names.
    :param optimizers: An optimizer name or list of optimizer names.
    :param training_loops: A training loop name or list of training loop names.
    :param epochs: A quick way to set the ``num_epochs`` in the training kwargs.
    :param create_inverse_triples: Either a boolean for a single entry or a list of booleans.
    :param regularizers: A regularizer name, list of regularizer names, or None if no regularizer is desired.
    :param negative_sampler: A negative sampler name, list of regularizer names, or None if no negative sampler
        is desired. Negative sampling is used only in combination with :class:`pykeen.training.SLCWATrainingLoop`.
    :param evaluator: The name of the evaluator to be used. Defaults to rank-based evaluator.
    :param stopper: The name of the stopper to be used. Defaults to NopStopper which doesn't define a
        stopping criterion.
    :param model_to_model_kwargs: A mapping from model name to dictionaries of default keyword arguments for
        the instantiation of that model.
    :param model_to_model_kwargs_ranges: A mapping from model name to dictionaries of keyword argument
        ranges for that model to be used in HPO.
    :param model_to_loss_to_loss_kwargs: A mapping from model name to a mapping of loss name to a mapping
        of default keyword arguments for the instantiation of that loss function. This is useful because for some
        losses, have hyper-parameters such as :class:`pykeen.losses.MarginRankingLoss`.
    :param model_to_loss_to_loss_kwargs_ranges: A mapping from model name to a mapping of loss name
        to a mapping of keyword argument ranges for that loss to be used in HPO.
    :param model_to_optimizer_to_optimizer_kwargs: A mapping from model name to a mapping of optimizer name to a mapping
        of default keyword arguments for the instantiation of that optimizer. This is useful because the optimizers,
        have hyper-parameters such as the learning rate.
    :param model_to_optimizer_to_optimizer_kwargs_ranges: A mapping from model name to a mapping of optimizer name
        to a mapping of keyword argument ranges for that optimizer to be used in HPO.
    :param model_to_regularizer_to_regularizer_kwargs: A mapping from model name to a mapping of regularizer name to a
        mapping of default keyword arguments for the instantiation of that regularizer. This is useful because the
        optimizers, have hyper-parameters such as the regularization weight.
    :param model_to_regularizer_to_regularizer_kwargs_ranges: A mapping from model name to a mapping of regularizer name
        to a mapping of keyword argument ranges for that regularizer to be used in HPO.
    :param model_to_negative_sampler_to_negative_sampler_kwargs: A mapping from model name to a mapping of
        negative sampler name to a mapping of default keyword arguments for the instantiation of that negative sampler.
        This is useful because the negative samplers, have hyper-parameters such as the number of negatives that should
        get generated for each positive training example.
    :param model_to_negative_sampler_to_negative_sampler_kwargs_ranges: A mapping from model name to a mapping of
        negative sampler name to a mapping of keyword argument ranges for that negative sampler to be used in HPO.
    :param model_to_training_loop_to_training_loop_kwargs: A mapping from model name to a mapping of training loop name
        to a mapping of default keyword arguments for the training loop.
    :param model_to_training_loop_to_training_kwargs: A mapping from model name to a mapping of trainer name to a
        mapping of default keyword arguments for the training procedure. This is useful because you can set the
        hyper-parameters such as the number of training epochs and the batch size.
    :param model_to_training_loop_to_training_kwargs_ranges:  A mapping from model name to a mapping of
        trainer name to a mapping of keyword argument ranges for that trainer to be used in HPO.
    :param evaluator_kwargs: The keyword arguments passed to the evaluator.
    :param evaluation_kwargs: The keyword arguments passed during evaluation.
    :param stopper_kwargs: The keyword arguments passed to the stopper.
    :param n_trials: Number of HPO trials.
    :param timeout: The time (seconds) after which the ablation study will be terminated.
    :param metric: The metric to optimize during HPO.
    :param direction: Defines, whether to 'maximize' or 'minimize' the metric during HPO.
    :param sampler: The HPO sampler, it defaults to random search.
    :param pruner: Defines approach for pruning trials. Per default no pruning is used, i.e., pruner is
        set to 'Nopruner'.
    :param metadata: A mapping of meta data arguments such as name of the ablation study.
    :param save_artifacts: Defines, whether each trained model sampled during HPO should be saved.
    :param move_to_cpu: Defines, whether a replicate of the best model should be moved to CPU.
    :param dry_run: Defines whether only the configurations for the single experiments should be created without
        running them.
    :param best_replicates: Defines how often the final model should be re-trained and evaluated based on the best
        hyper-parameters enabling to measure the variance in performance.
    :param discard_replicates: Defines, whether the best model should be discarded after training and evaluation.
    :param create_unique_subdir: Defines, whether a unique sub-directory for the experimental artifacts should
        be created. The sub-directory name is defined  by the  current  data + a unique id.
    """
    if isinstance(directory, str):
        directory = pathlib.Path(directory).resolve()
    if create_unique_subdir:
        directory = _create_path_with_id(directory=directory)

    directories = prepare_ablation(
        datasets=datasets,
        models=models,
        losses=losses,
        optimizers=optimizers,
        training_loops=training_loops,
        epochs=epochs,
        create_inverse_triples=create_inverse_triples,
        regularizers=regularizers,
        model_to_model_kwargs=model_to_model_kwargs,
        model_to_model_kwargs_ranges=model_to_model_kwargs_ranges,
        model_to_loss_to_loss_kwargs=model_to_loss_to_loss_kwargs,
        model_to_loss_to_loss_kwargs_ranges=model_to_loss_to_loss_kwargs_ranges,
        model_to_optimizer_to_optimizer_kwargs=model_to_optimizer_to_optimizer_kwargs,
        model_to_optimizer_to_optimizer_kwargs_ranges=model_to_optimizer_to_optimizer_kwargs_ranges,
        negative_sampler=negative_sampler,
        model_to_neg_sampler_to_neg_sampler_kwargs=model_to_negative_sampler_to_negative_sampler_kwargs,
        model_to_neg_sampler_to_neg_sampler_kwargs_ranges=model_to_negative_sampler_to_negative_sampler_kwargs_ranges,
        model_to_training_loop_to_training_loop_kwargs=model_to_training_loop_to_training_loop_kwargs,
        model_to_training_loop_to_training_kwargs=model_to_training_loop_to_training_kwargs,
        model_to_training_loop_to_training_kwargs_ranges=model_to_training_loop_to_training_kwargs_ranges,
        model_to_regularizer_to_regularizer_kwargs=model_to_regularizer_to_regularizer_kwargs,
        model_to_regularizer_to_regularizer_kwargs_ranges=model_to_regularizer_to_regularizer_kwargs_ranges,
        evaluator=evaluator,
        n_trials=n_trials,
        timeout=timeout,
        metric=metric,
        direction=direction,
        sampler=sampler,
        pruner=pruner,
        evaluator_kwargs=evaluator_kwargs,
        evaluation_kwargs=evaluation_kwargs,
        stopper=stopper,
        stopper_kwargs=stopper_kwargs,
        metadata=metadata,
        directory=directory,
        save_artifacts=save_artifacts,
    )

    _run_ablation_experiments(
        directories=directories,
        best_replicates=best_replicates,
        dry_run=dry_run,
        move_to_cpu=move_to_cpu,
        discard_replicates=discard_replicates,
    )


def _run_ablation_experiments(
    directories: Sequence[Tuple[Union[str, pathlib.Path], Union[str, pathlib.Path]]],
    best_replicates: Optional[int] = None,
    dry_run: bool = False,
    move_to_cpu: bool = True,
    discard_replicates: bool = False,
) -> None:
    """Run ablation experiments."""
    if dry_run:
        return

    from pykeen.hpo import hpo_pipeline_from_path

    for output_directory, rv_config_path in directories:
        if isinstance(output_directory, str):
            output_directory = pathlib.Path(output_directory).resolve()
        hpo_pipeline_result = hpo_pipeline_from_path(rv_config_path)
        hpo_pipeline_result.save_to_directory(output_directory)

        if not best_replicates:
            continue

        best_pipeline_dir = output_directory.joinpath('best_pipeline')
        best_pipeline_dir.mkdir(exist_ok=True, parents=True)
        logger.info('Re-training best pipeline and saving artifacts in %s', best_pipeline_dir)
        hpo_pipeline_result.replicate_best_pipeline(
            replicates=best_replicates,
            move_to_cpu=move_to_cpu,
            save_replicates=not discard_replicates,
            directory=best_pipeline_dir,
        )


def _create_path_with_id(directory: pathlib.Path) -> pathlib.Path:
    """Add unique id to path."""
    datetime = time.strftime('%Y-%m-%d-%H-%M')
    return directory.joinpath(f'{datetime}_{uuid4()}')


def ablation_pipeline_from_config(
    config: Mapping[str, Any],
    directory: str,
    *,
    dry_run: bool = False,
    best_replicates: Optional[int] = None,
    save_artifacts: bool = True,
    move_to_cpu: bool = True,
    discard_replicates: bool = False,
):
    """Generate a set of HPO configurations.

    A sample file can be run with``pykeen experiment ablation tests/resources/hpo_complex_nations.json``.

    :param config: Dictionary defining the ablation studies.
    :param directory: The directory in which the experimental artifacts will be saved.
    :param dry_run: Defines whether only the configurations for the single experiments should be created without
        running them.
    :param best_replicates: Defines how often the final model should be re-trained and evaluated based on the best
        hyper-parameters enabling to measure the variance in performance.
    :param save_artifacts: Defines, whether each trained model sampled during HPO should be saved.
    :param move_to_cpu: Defines, whether a replicate of the best model should be moved to CPU.
        We recommend to set this flag to 'True' to avoid unnecessary GPU usage.
    :param discard_replicates: Defines, whether the best model should be discarded after training and evaluation.
    :return: None.
    """
    return ablation_pipeline(
        **config,
        directory=directory,
        dry_run=dry_run,
        best_replicates=best_replicates,
        save_artifacts=save_artifacts,
        move_to_cpu=move_to_cpu,
        discard_replicates=discard_replicates,
    )


def prepare_ablation_from_path(
    path: Union[str, pathlib.Path],
    directory: Union[str, pathlib.Path],
    save_artifacts: bool,
) -> List[Tuple[pathlib.Path, pathlib.Path]]:
    """Prepare a set of ablation study directories.

    :param path: Path to configuration file defining the ablation studies.
    :param directory: The directory in which the experimental artifacts (including the ablation configurations)
        will be saved.
    :param save_artifacts: Defines, whether the output directories for the trained models sampled during HPO should be
        created.
    :return: pairs of output directories and HPO config paths inside those directories
    """
    if isinstance(directory, str):
        directory = pathlib.Path(directory).resolve()
    directory = _create_path_with_id(directory=directory)
    with open(path) as file:
        config = json.load(file)
    return prepare_ablation_from_config(config=config, directory=directory, save_artifacts=save_artifacts)


def prepare_ablation_from_config(
    config: Mapping[str, Any],
    directory: Union[str, pathlib.Path],
    save_artifacts: bool,
) -> List[Tuple[pathlib.Path, pathlib.Path]]:
    """Prepare a set of ablation study directories.

    :param config: Dictionary defining the ablation studies.
    :param directory: The directory in which the experimental artifacts (including the ablation configurations)
        will be saved.
    :param save_artifacts: Defines, whether the output directories for the trained models sampled during HPO should be
        created.

    :return: pairs of output directories and HPO config paths inside those directories
    """
    metadata = config['metadata']
    optuna_config = config['optuna']
    ablation_config = config['ablation']

    return prepare_ablation(
        **ablation_config,
        **optuna_config,
        metadata=metadata,
        directory=directory,
        save_artifacts=save_artifacts,
    )


def prepare_ablation(  # noqa:C901
    datasets: Union[str, List[str]],
    models: Union[str, List[str]],
    losses: Union[str, List[str]],
    optimizers: Union[str, List[str]],
    training_loops: Union[str, List[str]],
    directory: Union[str, pathlib.Path],
    *,
    epochs: Optional[int] = None,
    create_inverse_triples: Union[bool, List[bool]] = False,
    regularizers: Union[None, str, List[str], List[None]] = None,
    negative_sampler: Optional[str] = None,
    evaluator: Optional[str] = None,
    model_to_model_kwargs: Optional[Mapping2D] = None,
    model_to_model_kwargs_ranges: Optional[Mapping2D] = None,
    model_to_loss_to_loss_kwargs: Optional[Mapping3D] = None,
    model_to_loss_to_loss_kwargs_ranges: Optional[Mapping3D] = None,
    model_to_optimizer_to_optimizer_kwargs: Optional[Mapping3D] = None,
    model_to_optimizer_to_optimizer_kwargs_ranges: Optional[Mapping3D] = None,
    model_to_training_loop_to_training_loop_kwargs: Optional[Mapping3D] = None,
    model_to_neg_sampler_to_neg_sampler_kwargs: Optional[Mapping3D] = None,
    model_to_neg_sampler_to_neg_sampler_kwargs_ranges: Optional[Mapping3D] = None,
    model_to_training_loop_to_training_kwargs: Optional[Mapping3D] = None,
    model_to_training_loop_to_training_kwargs_ranges: Optional[Mapping3D] = None,
    model_to_regularizer_to_regularizer_kwargs: Optional[Mapping3D] = None,
    model_to_regularizer_to_regularizer_kwargs_ranges: Optional[Mapping3D] = None,
    n_trials: Optional[int] = 5,
    timeout: Optional[int] = 3600,
    metric: Optional[str] = 'hits@10',
    direction: Optional[str] = 'maximize',
    sampler: Optional[str] = 'random',
    pruner: Optional[str] = 'nop',
    evaluator_kwargs: Optional[Mapping[str, Any]] = None,
    evaluation_kwargs: Optional[Mapping[str, Any]] = None,
    stopper: Optional[str] = 'NopStopper',
    stopper_kwargs: Optional[Mapping[str, Any]] = None,
    metadata: Optional[Mapping] = None,
    save_artifacts: bool = True,
) -> List[Tuple[pathlib.Path, pathlib.Path]]:
    """Prepare an ablation directory.

    :param datasets: A dataset name or list of dataset names.
    :param models: A model name or list of model names.
    :param losses: A loss function name or list of loss function names.
    :param optimizers: An optimizer name or list of optimizer names.
    :param training_loops: A training loop name or list of training loop names.
    :param epochs: A quick way to set the ``num_epochs`` in the training kwargs.
    :param create_inverse_triples: Either a boolean for a single entry or a list of booleans.
    :param regularizers: A regularizer name, list of regularizer names, or None if no regularizer is desired.
    :param negative_sampler: A negative sampler name, list of regularizer names, or None if no negative sampler
        is desired. Negative sampling is used only in combination with the pykeen.training.sclwa training loop.
    :param evaluator: The name of the evaluator to be used. Defaults to rank-based evaluator.
    :param stopper: The name of the stopper to be used. Defaults to NopStopper which doesn't define a
        stopping criterion.
    :param model_to_model_kwargs: A mapping from model name to dictionaries of default keyword arguments for
        the instantiation of that model.
    :param model_to_model_kwargs_ranges: A mapping from model name to dictionaries of keyword argument
        ranges for that model to be used in HPO.
    :param model_to_loss_to_loss_kwargs: A mapping from model name to a mapping of loss name to a mapping
        of default keyword arguments for the instantiation of that loss function. This is useful because for some
        losses, have hyper-parameters such as pykeen.losses.MarginRankingLoss
    :param model_to_loss_to_loss_kwargs_ranges: A mapping from model name to a mapping of loss name
        to a mapping of keyword argument ranges for that loss to be used in HPO.
    :param model_to_optimizer_to_optimizer_kwargs: A mapping from model name to a mapping of optimizer name to a mapping
        of default keyword arguments for the instantiation of that optimizer. This is useful because the optimizers,
        have hyper-parameters such as the learning rate.
    :param model_to_optimizer_to_optimizer_kwargs_ranges: A mapping from model name to a mapping of optimizer name
        to a mapping of keyword argument ranges for that optimizer to be used in HPO.
    :param model_to_regularizer_to_regularizer_kwargs: A mapping from model name to a mapping of regularizer name to a
        mapping of default keyword arguments for the instantiation of that regularizer. This is useful because the
        optimizers, have hyper-parameters such as the regularization weight.
    :param model_to_regularizer_to_regularizer_kwargs_ranges: A mapping from model name to a mapping of regularizer name
        to a mapping of keyword argument ranges for that regularizer to be used in HPO.
    :param model_to_neg_sampler_to_neg_sampler_kwargs: A mapping from model name to a mapping of
        negative sampler name to a mapping of default keyword arguments for the instantiation of that negative sampler.
        This is useful because the negative samplers, have hyper-parameters such as the number of negatives that should
        get generated for each positive training example.
    :param model_to_neg_sampler_to_neg_sampler_kwargs_ranges: A mapping from model name to a mapping of
        negative sampler name to a mapping of keyword argument ranges for that negative sampler to be used in HPO.
    :param model_to_training_loop_to_training_loop_kwargs: A mapping from model name to a mapping of training loop name
        to a mapping of default keyword arguments for the training loop.
    :param model_to_training_loop_to_training_kwargs: A mapping from model name to a mapping of trainer name to a
        mapping of default keyword arguments for the training procedure. This is useful because you can set the
        hyper-parameters such as the number of training epochs and the batch size.
    :param model_to_training_loop_to_training_kwargs_ranges:  A mapping from model name to a mapping of
        trainer name to a mapping of keyword argument ranges for that trainer to be used in HPO.
    :param evaluator_kwargs: The keyword arguments passed to the evaluator.
    :param evaluation_kwargs: The keyword arguments passed during evaluation.
    :param stopper_kwargs: The keyword arguments passed to the stopper.
    :param n_trials: Number of HPO trials.
    :param timeout: The time (seconds) after which the ablation study will be terminated.
    :param metric: The metric to optimize during HPO.
    :param direction: Defines, whether to 'maximize' or 'minimize' the metric during HPO.
    :param sampler: The HPO sampler, it defaults to random search.
    :param pruner: Defines approach for pruning trials. Per default no pruning is used, i.e., pruner is
        set to 'Nopruner'.
    :param metadata: A mapping of meta data arguments such as name of the ablation study.
    :param directory: The directory in which the experimental artifacts will be saved.
    :param save_artifacts: Defines, whether each trained model sampled during HPO should be saved.

    :return: pairs of output directories and HPO config paths inside those directories.
    :raises ValueError:
            If the dataset is not specified correctly, i.e., dataset is not of type str, or a dictionary containing
            the paths to the training, testing, and validation data.
    """
    if isinstance(directory, str):
        directory = pathlib.Path(directory).resolve()
    if isinstance(datasets, str):
        datasets = [datasets]
    if isinstance(create_inverse_triples, bool):
        create_inverse_triples = [create_inverse_triples]
    if isinstance(models, str):
        models = [models]
    if isinstance(losses, str):
        losses = [losses]
    if isinstance(optimizers, str):
        optimizers = [optimizers]
    if isinstance(training_loops, str):
        training_loops = [training_loops]
    if isinstance(regularizers, str):
        regularizers = [regularizers]
    elif regularizers is None:
        regularizers = [None]

    it = itt.product(
        datasets,
        create_inverse_triples,
        models,
        losses,
        regularizers,
        optimizers,
        training_loops,
    )

    if not model_to_model_kwargs:
        model_to_model_kwargs = {}
    if not model_to_model_kwargs_ranges:
        model_to_model_kwargs_ranges = {}

    directories = []
    for counter, (
        dataset,
        create_inverse_triples,
        model,
        loss,
        regularizer,
        optimizer,
        training_loop,
    ) in enumerate(it):
        dataset_name = normalize_string(dataset) if isinstance(dataset, str) else 'user_data'
        experiment_name = f'{counter:04d}_{dataset_name}_{normalize_string(model)}'
        output_directory = directory.joinpath(experiment_name)
        output_directory.mkdir(exist_ok=True, parents=True)
        # TODO what happens if already exists?

        _experiment_optuna_config = {
            'n_trials': n_trials,
            'timeout': timeout,
            'metric': metric,
            'direction': direction,
            'sampler': sampler,
            'pruner': pruner,
        }
        _experiment_optuna_config['storage'] = f'sqlite:///{output_directory.as_posix()}/optuna_results.db'
        if save_artifacts:
            save_model_directory = output_directory.joinpath('artifacts')
            save_model_directory.mkdir(exist_ok=True, parents=True)
            _experiment_optuna_config['save_model_directory'] = save_model_directory.as_posix()

        hpo_config: Dict[str, Any] = dict()
        hpo_config['stopper'] = stopper

        if stopper_kwargs is not None:
            hpo_config['stopper_kwargs'] = stopper_kwargs

        # TODO incorporate setting of random seed
        # pipeline_kwargs=dict(
        #    random_seed=random_non_negative_int(),
        # ),

        def _set_arguments(config: Optional[Mapping3D], key: str, value: str) -> None:
            """Set argument and its values."""
            d = {}
            d[key] = {} if config is None else config.get(model, {}).get(value, {})
            if d[key]:
                hpo_config.update(d)

        # Add dataset to current_pipeline
        if isinstance(dataset, str):
            hpo_config['dataset'] = dataset
        elif isinstance(dataset, dict):
            # Training, test, and validation paths are provided
            hpo_config['training'] = dataset['training']
            hpo_config['testing'] = dataset['testing']
            hpo_config['validation'] = dataset['validation']
        else:
            raise ValueError(
                "Dataset must be either the dataset name, i.e., of type str, or a dictionary containing\n"
                "the paths to the training, testing, and validation data.",
            )
        logger.info(f"Dataset: {dataset}")
        hpo_config['dataset_kwargs'] = dict(create_inverse_triples=create_inverse_triples)
        logger.info(f"Add inverse triples: {create_inverse_triples}")

        hpo_config['model'] = model
        hpo_config['model_kwargs'] = model_to_model_kwargs.get(model, {})
        hpo_config['model_kwargs_ranges'] = model_to_model_kwargs_ranges.get(model, {})
        logger.info(f"Model: {model}")

        # Add loss function to current_pipeline
        hpo_config['loss'] = loss
        _set_arguments(config=model_to_loss_to_loss_kwargs, key='loss_kwargs', value=loss)
        _set_arguments(config=model_to_loss_to_loss_kwargs_ranges, key='loss_kwargs_ranges', value=loss)
        logger.info(f"Loss functions: {loss}")

        # Add regularizer to current_pipeline
        if regularizer is not None:
            hpo_config['regularizer'] = regularizer
            _set_arguments(
                config=model_to_regularizer_to_regularizer_kwargs,
                key='regularizer_kwargs',
                value=regularizer,
            )
            _set_arguments(
                config=model_to_regularizer_to_regularizer_kwargs_ranges,
                key='regularizer_kwargs_ranges',
                value=regularizer,
            )
            logger.info(f"Regularizer: {regularizer}")

        # Add optimizer to current_pipeline
        hpo_config['optimizer'] = optimizer
        _set_arguments(config=model_to_optimizer_to_optimizer_kwargs, key='optimizer_kwargs', value=optimizer)
        _set_arguments(
            config=model_to_optimizer_to_optimizer_kwargs_ranges,
            key='optimizer_kwargs_ranges',
            value=optimizer,
        )
        logger.info(f"Optimizer: {optimizer}")

        # Add training approach to current_pipeline
        hpo_config['training_loop'] = training_loop
        _set_arguments(
            config=model_to_training_loop_to_training_loop_kwargs,
            key='training_loop_kwargs',
            value=training_loop,
        )
        _set_arguments(config=model_to_training_loop_to_training_kwargs, key='training_kwargs', value=training_loop)
        _set_arguments(
            config=model_to_training_loop_to_training_kwargs_ranges,
            key='training_kwargs_ranges',
            value=training_loop,
        )
        logger.info(f"Training loop: {training_loop}")

        if normalize_string(training_loop, suffix=_TRAINING_LOOP_SUFFIX) == 'slcwa':
            negative_sampler = negative_sampler or 'basic'  # default to basic
            _set_arguments(
                config=model_to_neg_sampler_to_neg_sampler_kwargs,
                key='negative_sampler_kwargs',
                value=negative_sampler,
            )
            _set_arguments(
                config=model_to_neg_sampler_to_neg_sampler_kwargs_ranges,
                key='negative_sampler_kwargs_ranges',
                value=negative_sampler,
            )
            logger.info(f"Negative sampler: {negative_sampler}")

        # Add evaluation
        hpo_config['evaluator'] = evaluator
        if evaluator_kwargs:
            hpo_config['evaluator_kwargs'] = evaluator_kwargs
        hpo_config['evaluation_kwargs'] = evaluation_kwargs or {}
        logger.info(f"Evaluator: {evaluator}")

        if epochs is not None:
            hpo_config.setdefault('training_kwargs', {}).setdefault('num_epochs', epochs)

        rv_config = dict(
            type='hpo',
            metadata=metadata or {},
            pipeline=hpo_config,
            optuna=_experiment_optuna_config,
        )

        rv_config_path = output_directory.joinpath('hpo_config.json')
        with rv_config_path.open('w') as file:
            json.dump(rv_config, file, indent=2, ensure_ascii=True)

        directories.append((output_directory, rv_config_path))

    return directories



# -*- coding: utf-8 -*-

"""Tests for deterioration workflow."""

import unittest

from pykeen.datasets.base import Dataset
from pykeen.datasets.nations import Nations
from pykeen.triples.triples_factory import splits_steps
from pykeen.utils import ensure_torch_random_state


class TestDeterioration(unittest.TestCase):
    """Tests for deterioration workflow."""

    def setUp(self) -> None:
        """Set up deterioration tests."""
        self.generator = ensure_torch_random_state(42)
        self.reference = Nations()
        self.num_training_triples = self.reference.training.num_triples
        self.num_triples = (
            self.reference.training.num_triples
            + self.reference.testing.num_triples
            + self.reference.validation.num_triples
        )

    def test_deteriorate(self):
        """Test deterioration on integer values for ``n``."""
        for n in [1, 2, 5, 10, 50, 100, 500, 1000]:
            with self.subTest(n=n):
                derived = self.reference.deteriorate(n=n, random_state=self.generator)
                self._help_check(derived)
                self.assertEqual(n, splits_steps(self.reference._tup(), derived._tup()))
                self.assertEqual(1 - n / self.num_triples, self.reference.similarity(derived), msg='similarity')
                self.assertEqual(1 - n / self.num_triples, derived.similarity(self.reference), msg='similarity')

    def test_deteriorate_frac(self):
        """Test deterioration on fractional values for ``n``."""
        for n_frac in [
            1 / self.num_training_triples,
            2 / self.num_training_triples,
            5 / self.num_training_triples,
            0.1,
            0.2,
            0.3,
        ]:
            n = int(n_frac * self.num_training_triples)
            with self.subTest(n=n, n_frac=n_frac):
                derived = self.reference.deteriorate(n=n, random_state=self.generator)
                self._help_check(derived)
                self.assertEqual(
                    n,
                    splits_steps(self.reference._tup(), derived._tup()),
                    msg='steps',
                )
                self.assertEqual(1 - n / self.num_triples, self.reference.similarity(derived), msg='similarity')
                self.assertEqual(1 - n / self.num_triples, derived.similarity(self.reference), msg='similarity')

    def _help_check(self, derived: Dataset):
        self.assertIsNotNone(derived.validation)
        self.assertEqual(self.num_training_triples, self.reference.training.num_triples)
        self.assertEqual(
            self.num_triples,
            sum((
                derived.training.num_triples,
                derived.testing.num_triples,
                derived.validation.num_triples,
            )),
            msg='different number of total triples',
        )
        self.assertLess(derived.training.num_triples, self.reference.training.num_triples)



# -*- coding: utf-8 -*-

"""Tests for dataset analysis utilities."""

import itertools
import unittest
from typing import Iterable, Mapping

import numpy as np
import pandas

from pykeen.datasets import Dataset, Nations, analysis as dataset_analysis
from pykeen.triples import analysis as triple_analysis


def _old_skyline(xs):
    # naive implementation, O(n2)
    return {
        (s, c)
        for s, c in xs
        if not any(
            s2 >= s and c2 >= c
            for s2, c2 in xs
            if (s, c) != (s2, c2)
        )
    }


class TestUtils(unittest.TestCase):
    """Test skyline."""

    def test_skyline(self):
        """Test the skyline function."""
        n = 500
        pairs = list(zip(
            np.random.randint(low=0, high=200, size=n, dtype=int),
            np.random.uniform(0, 6, size=n),
        ))
        self.assertEqual(set(_old_skyline(pairs)), set(triple_analysis._get_skyline(pairs)))


def _test_count_dataframe(
    dataset: Dataset,
    df: pandas.DataFrame,
    labels: bool = True,
    merge_subsets: bool = True,
    merge_sides: bool = True,
):
    """Check the general structure of a count dataframe."""
    # check correct output type
    assert isinstance(df, pandas.DataFrame)

    expected_columns = {triple_analysis.COUNT_COLUMN_NAME}
    expected_columns.update(_check_labels(
        df=df,
        labels=labels,
        id_column_name=triple_analysis.ENTITY_ID_COLUMN_NAME,
        label_column_name=triple_analysis.ENTITY_LABEL_COLUMN_NAME,
        label_to_id=dataset.entity_to_id,
    ))
    expected_columns.update(_check_labels(
        df=df,
        labels=labels,
        id_column_name=triple_analysis.RELATION_ID_COLUMN_NAME,
        label_column_name=triple_analysis.RELATION_LABEL_COLUMN_NAME,
        label_to_id=dataset.relation_to_id,
    ))

    if not merge_subsets:
        expected_columns.add(dataset_analysis.SUBSET_COLUMN_NAME)

        # check value range subset
        assert df[dataset_analysis.SUBSET_COLUMN_NAME].isin(dataset.factory_dict.keys()).all()

    if not merge_sides:
        expected_columns.add(triple_analysis.ENTITY_POSITION_COLUMN_NAME)

        # check value range side
        assert df[triple_analysis.ENTITY_POSITION_COLUMN_NAME].isin({
            triple_analysis.POSITION_HEAD,
            triple_analysis.POSITION_TAIL,
        }).all()

    # check columns
    assert expected_columns == set(df.columns)

    # check value range and type
    assert (df[triple_analysis.COUNT_COLUMN_NAME] >= 0).all()
    assert df[triple_analysis.COUNT_COLUMN_NAME].dtype == np.int64


def _check_labels(
    df: pandas.DataFrame,
    labels: bool,
    id_column_name: str,
    label_column_name: str,
    label_to_id: Mapping[str, int],
) -> Iterable[str]:
    if id_column_name in df.columns:
        yield id_column_name

        # check value range entity IDs
        assert df[id_column_name].isin(label_to_id.values()).all()

        if labels:
            yield label_column_name

            # check value range entity labels
            assert df[label_column_name].isin(label_to_id.keys()).all()


class DatasetAnalysisTests(unittest.TestCase):
    """Tests for dataset analysis utilities."""

    def setUp(self) -> None:
        """Initialize the unittest."""
        self.dataset = Nations()

    def test_relation_count_dataframe(self):
        """Test relation count dataframe."""
        for labels, merge_subsets in itertools.product((False, True), repeat=2):
            _test_count_dataframe(
                dataset=self.dataset,
                df=dataset_analysis.get_relation_count_df(
                    dataset=self.dataset,
                    add_labels=labels,
                    merge_subsets=merge_subsets,
                ),
                labels=labels,
                merge_subsets=merge_subsets,
            )

    def test_entity_count_dataframe(self):
        """Test entity count dataframe."""
        for labels, merge_subsets, merge_sides in itertools.product((False, True), repeat=3):
            _test_count_dataframe(
                dataset=self.dataset,
                df=dataset_analysis.get_entity_count_df(
                    dataset=self.dataset,
                    add_labels=labels,
                    merge_subsets=merge_subsets,
                    merge_sides=merge_sides,
                ),
                labels=labels,
                merge_subsets=merge_subsets,
                merge_sides=merge_sides,
            )

    def test_entity_relation_co_occurrence_dataframe(self):
        """Test entity-relation co-occurrence dataframe."""
        for labels, merge_sides, merge_subsets in itertools.product((False, True), repeat=3):
            _test_count_dataframe(
                dataset=self.dataset,
                df=dataset_analysis.get_entity_relation_co_occurrence_df(
                    dataset=self.dataset,
                    merge_sides=merge_sides,
                    merge_subsets=merge_subsets,
                    add_labels=labels,
                ),
                labels=labels,
                merge_subsets=merge_subsets,
                merge_sides=merge_sides,
            )

    def test_relation_pattern_types(self):
        """Helper method for relation pattern classification."""
        df = dataset_analysis.get_relation_pattern_types_df(
            dataset=self.dataset,
            drop_confidence=False,
        )

        # check correct type
        assert isinstance(df, pandas.DataFrame)

        # check relation_id value range
        assert df[triple_analysis.RELATION_ID_COLUMN_NAME].isin(self.dataset.relation_to_id.values()).all()

        # check pattern value range
        assert df[triple_analysis.PATTERN_TYPE_COLUMN_NAME].isin(triple_analysis.RELATION_PATTERN_TYPES).all()

        # check confidence value range
        x = df[triple_analysis.CONFIDENCE_COLUMN_NAME].values
        assert (0 <= x).all()
        assert (x <= 1).all()

        # check support value range
        x = df[triple_analysis.SUPPORT_COLUMN_NAME].values
        assert (1 <= x).all()

    def test_relation_cardinality_types(self):
        """Tests for relation cardinality type classification."""
        df = dataset_analysis.get_relation_cardinality_types_df(
            dataset=self.dataset,
        )

        # check correct type
        assert isinstance(df, pandas.DataFrame)

        # check relation_id value range
        assert df[triple_analysis.RELATION_ID_COLUMN_NAME].isin(self.dataset.relation_to_id.values()).all()

        # check pattern value range
        assert df[triple_analysis.CARDINALITY_TYPE_COLUMN_NAME].isin(triple_analysis.RELATION_CARDINALITY_TYPES).all()

    def test_calculate_relation_functionality(self):
        """Tests calculate_relation_functionality."""
        df = dataset_analysis.get_relation_functionality_df(
            dataset=self.dataset,
        )

        # check correct type
        assert isinstance(df, pandas.DataFrame)

        assert {
            triple_analysis.RELATION_ID_COLUMN_NAME,
            triple_analysis.FUNCTIONALITY_COLUMN_NAME,
            triple_analysis.INVERSE_FUNCTIONALITY_COLUMN_NAME,
        }.issubset(df.columns)

        # check relation_id value range
        assert df[triple_analysis.RELATION_ID_COLUMN_NAME].isin(self.dataset.relation_to_id.values()).all()



# -*- coding: utf-8 -*-

"""Test hyper-parameter optimization."""

import tempfile
import unittest

import optuna
import pytest

from pykeen.datasets.nations import (
    NATIONS_TEST_PATH, NATIONS_TRAIN_PATH, NATIONS_VALIDATE_PATH, Nations,
    NationsLiteral,
)
from pykeen.hpo import hpo_pipeline
from pykeen.hpo.hpo import suggest_kwargs
from pykeen.triples import TriplesFactory


class TestInvalidConfigurations(unittest.TestCase):
    """Tests of invalid HPO configurations."""

    def test_earl_stopping_with_optimize_epochs(self):
        """Assert that the pipeline raises a value error."""
        with self.assertRaises(ValueError):
            hpo_pipeline(
                dataset='kinships',
                model='transe',
                stopper='early',
                training_kwargs_ranges=dict(epochs=...),
            )


@pytest.mark.slow
class TestHyperparameterOptimization(unittest.TestCase):
    """Test hyper-parameter optimization."""

    def test_run(self):
        """Test simply making a study."""
        hpo_pipeline_result = hpo_pipeline(
            dataset='nations',
            model='TransE',
            training_kwargs=dict(num_epochs=5, use_tqdm=False),
            n_trials=2,
        )
        df = hpo_pipeline_result.study.trials_dataframe(multi_index=True)
        # Check a model param is optimized
        self.assertIn(('params', 'model.embedding_dim'), df.columns)
        # Check a loss param is optimized
        self.assertIn(('params', 'loss.margin'), df.columns)
        self.assertNotIn(('params', 'training.num_epochs'), df.columns)

    def test_specified_model_hyperparameter(self):
        """Test making a study that has a specified model hyper-parameter."""
        target_embedding_dim = 50
        hpo_pipeline_result = hpo_pipeline(
            dataset='nations',
            model='TransE',
            model_kwargs=dict(embedding_dim=target_embedding_dim),
            training_kwargs=dict(num_epochs=5, use_tqdm=False),
            n_trials=2,
        )
        df = hpo_pipeline_result.study.trials_dataframe(multi_index=True)
        # Check a model param is NOT optimized
        self.assertNotIn(('params', 'model.embedding_dim'), df.columns)
        # Check a loss param is optimized
        self.assertIn(('params', 'loss.margin'), df.columns)

    def test_specified_loss_hyperparameter(self):
        """Test making a study that has a specified loss hyper-parameter."""
        hpo_pipeline_result = hpo_pipeline(
            dataset='nations',
            model='TransE',
            loss_kwargs=dict(margin=1.0),
            training_kwargs=dict(num_epochs=5, use_tqdm=False),
            n_trials=2,
        )
        df = hpo_pipeline_result.study.trials_dataframe(multi_index=True)
        # Check a model param is optimized
        self.assertIn(('params', 'model.embedding_dim'), df.columns)
        # Check a loss param is NOT optimized
        self.assertNotIn(('params', 'loss.margin'), df.columns)

    def test_specified_loss_and_model_hyperparameter(self):
        """Test making a study that has a specified loss hyper-parameter."""
        target_embedding_dim = 50
        hpo_pipeline_result = hpo_pipeline(
            dataset='nations',
            model='TransE',
            model_kwargs=dict(embedding_dim=target_embedding_dim),
            loss='MarginRankingLoss',
            loss_kwargs=dict(margin=1.0),
            training_kwargs=dict(num_epochs=5, use_tqdm=False),
            n_trials=2,
        )
        df = hpo_pipeline_result.study.trials_dataframe(multi_index=True)
        # Check a model param is NOT optimized
        self.assertNotIn(('params', 'model.embedding_dim'), df.columns)
        # Check a loss param is NOT optimized
        self.assertNotIn(('params', 'loss.margin'), df.columns)

    def test_specified_range(self):
        """Test making a study that has a specified hyper-parameter."""
        hpo_pipeline_result = hpo_pipeline(
            dataset='nations',
            model='TransE',
            model_kwargs_ranges=dict(
                embedding_dim=dict(type=int, low=60, high=80, q=10),
            ),
            loss_kwargs_ranges=dict(
                margin=dict(type=int, low=1, high=2),
            ),
            training_kwargs=dict(num_epochs=5, use_tqdm=False),
            n_trials=2,
        )
        df = hpo_pipeline_result.study.trials_dataframe(multi_index=True)
        self.assertIn(('params', 'model.embedding_dim'), df.columns)
        self.assertTrue(df[('params', 'model.embedding_dim')].isin({60., 70., 80.}).all())

        self.assertIn(('params', 'loss.margin'), df.columns)
        self.assertTrue(df[('params', 'loss.margin')].isin({1, 2}).all())

    def test_sampling_values_from_2_power_x(self):
        """Test making a study that has a range defined by f(x) = 2^x."""
        model_kwargs_ranges = dict(
            embedding_dim=dict(type=int, low=0, high=4, scale='power_two'),
        )
        objective = _test_suggest(model_kwargs_ranges)
        study = optuna.create_study()
        study.optimize(objective, n_trials=2)

        df = study.trials_dataframe(multi_index=True)
        self.assertIn(('params', 'model.embedding_dim'), df.columns)
        self.assertTrue(df[('params', 'model.embedding_dim')].isin({1, 2, 4, 8, 16}).all())

        objective = _test_suggest(model_kwargs_ranges)
        with self.assertRaises(Exception) as context:
            study = optuna.create_study()
            study.optimize(objective, n_trials=2)
            self.assertIn('Upper bound 4 is not greater than lower bound 4.', context.exception)

    def test_sampling_values_from_power_x(self):
        """Test making a study that has a range defined by f(x) = base^x."""
        kwargs_ranges = dict(
            embedding_dim=dict(type=int, low=0, high=2, scale='power', base=10),
        )
        objective = _test_suggest(kwargs_ranges)
        study = optuna.create_study()
        study.optimize(objective, n_trials=2)

        df = study.trials_dataframe(multi_index=True)
        self.assertIn(('params', 'model.embedding_dim'), df.columns)
        values = df[('params', 'model.embedding_dim')]
        self.assertTrue(values.isin({1, 10, 100}).all(), msg=f'Got values: {values}')


def _test_suggest(kwargs_ranges):
    def objective(trial):
        suggest_kwargs(prefix='model', trial=trial, kwargs_ranges=kwargs_ranges)
        return 1.

    return objective


@pytest.mark.slow
class TestHPODatasets(unittest.TestCase):
    """Test different ways of loading data in HPO."""

    def test_custom_dataset_instance(self):
        """Test passing a pre-instantiated dataset to HPO."""
        hpo_pipeline_result = self._help_test_hpo(
            study_name='HPO with custom dataset instance',
            dataset=Nations(),  # mock a "custom" dataset by using one already available
        )
        # Since custom data was passed, we can't store any of this
        self.assertNotIn('dataset', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('training', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('testing', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('validation', hpo_pipeline_result.study.user_attrs)

    def test_custom_dataset_cls(self):
        """Test passing a dataset class to HPO."""
        hpo_pipeline_result = self._help_test_hpo(
            study_name='HPO with custom dataset class',
            dataset=Nations,
        )
        # currently, any custom data doesn't get stored.
        self.assertNotIn('dataset', hpo_pipeline_result.study.user_attrs)
        # self.assertEqual(Nations.get_normalized_name(), hpo_pipeline_result.study.user_attrs['dataset'])
        self.assertNotIn('training', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('testing', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('validation', hpo_pipeline_result.study.user_attrs)

    def test_custom_dataset_path(self):
        """Test passing a dataset class to HPO."""
        hpo_pipeline_result = self._help_test_hpo(
            study_name='HPO with custom dataset path',
            dataset=NATIONS_TRAIN_PATH,
        )
        self.assertIn('dataset', hpo_pipeline_result.study.user_attrs)
        self.assertEqual(str(NATIONS_TRAIN_PATH), hpo_pipeline_result.study.user_attrs['dataset'])
        self.assertNotIn('training', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('testing', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('validation', hpo_pipeline_result.study.user_attrs)

    def test_custom_tf_object(self):
        """Test using a custom triples factories with HPO.

        .. seealso:: https://github.com/pykeen/pykeen/issues/230
        """
        tf = TriplesFactory.from_path(path=NATIONS_TRAIN_PATH)
        training, testing, validation = tf.split([.8, .1, .1], random_state=0)

        hpo_pipeline_result = self._help_test_hpo(
            study_name='HPO with custom triples factories',
            training=training,
            testing=testing,
            validation=validation,
        )
        self.assertNotIn('dataset', hpo_pipeline_result.study.user_attrs)
        # Since there's no source path information, these shouldn't be
        # added, even if it might be possible to infer path information
        # from the triples factories
        self.assertNotIn('training', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('testing', hpo_pipeline_result.study.user_attrs)
        self.assertNotIn('validation', hpo_pipeline_result.study.user_attrs)

    def test_custom_paths(self):
        """Test using a custom triples paths with HPO."""
        hpo_pipeline_result = self._help_test_hpo(
            study_name='HPO with custom triples paths',
            training=NATIONS_TRAIN_PATH,  # mock "custom" paths
            testing=NATIONS_TEST_PATH,
            validation=NATIONS_VALIDATE_PATH,
        )
        self.assertNotIn('dataset', hpo_pipeline_result.study.user_attrs)
        # Since paths were passed for training, testing, and validation,
        # they should be stored as study-level attributes
        self.assertIn('training', hpo_pipeline_result.study.user_attrs)
        self.assertEqual(str(NATIONS_TRAIN_PATH), hpo_pipeline_result.study.user_attrs['training'])
        self.assertIn('testing', hpo_pipeline_result.study.user_attrs)
        self.assertEqual(str(NATIONS_TEST_PATH), hpo_pipeline_result.study.user_attrs['testing'])
        self.assertIn('validation', hpo_pipeline_result.study.user_attrs)
        self.assertEqual(str(NATIONS_VALIDATE_PATH), hpo_pipeline_result.study.user_attrs['validation'])

    def _help_test_hpo(self, **kwargs):
        hpo_pipeline_result = hpo_pipeline(
            **kwargs,
            model='TransE',
            n_trials=1,
            training_kwargs=dict(num_epochs=1, use_tqdm=False),
            evaluation_kwargs=dict(use_tqdm=False),
        )
        with tempfile.TemporaryDirectory() as directory:
            hpo_pipeline_result.save_to_directory(directory)
        return hpo_pipeline_result


@pytest.mark.slow
class TestHyperparameterOptimizationLiterals(unittest.TestCase):
    """Test hyper-parameter optimization."""

    def test_run(self):
        """Test simply making a study."""
        hpo_pipeline_result = hpo_pipeline(
            dataset=NationsLiteral,
            model='DistMultLiteral',
            training_kwargs=dict(num_epochs=5, use_tqdm=False),
            n_trials=2,
        )
        df = hpo_pipeline_result.study.trials_dataframe(multi_index=True)
        # Check a model param is optimized
        self.assertIn(('params', 'model.embedding_dim'), df.columns)
        # Check a loss param is optimized
        self.assertIn(('params', 'loss.margin'), df.columns)
        self.assertNotIn(('params', 'training.num_epochs'), df.columns)



# -*- coding: utf-8 -*-

"""Constants for PyKEEN testing."""

import pathlib

__all__ = [
    'HERE',
    'RESOURCES',
    'EPSILON',
]

HERE = pathlib.Path(__file__).resolve().parent
RESOURCES = HERE.joinpath('resources')
EPSILON = 1.0e-07



# -*- coding: utf-8 -*-

"""Unittest for training utilities."""

import unittest
from typing import Type

import numpy as np
import torch

from pykeen.models import Model, TransE
from pykeen.training.utils import lazy_compile_random_batches


class LossTensorTest(unittest.TestCase):
    """Test label smoothing."""

    model_cls: Type[Model] = TransE
    embedding_dim: int = 8

    def setUp(self):
        """Set up the loss tensor tests."""
        self.triples = np.array(
            [
                ['peter', 'likes', 'chocolate_cake'],
                ['chocolate_cake', 'isA', 'dish'],
                ['susan', 'likes', 'pizza'],
                ['peter', 'likes', 'susan'],
            ],
            dtype=str,
        )

        self.labels = torch.tensor([
            [0., 1., 0., 0., 0.],
            [0., 0., 0., 1., 0.],
            [1., 0., 0., 0., 1.],
        ])

        self.predictions = torch.tensor([
            [1., 0., 1., 1., 1.],
            [1., 1., 1., 0., 1.],
            [0., 1., 1., 1., 0.],
        ])


class BatchCompilationTest(unittest.TestCase):
    """Test compilation of random batches."""

    batch_size: int = 64
    num_samples: int = 256 + batch_size // 2  # to check whether the method works for incomplete batches
    num_entities: int = 10

    def setUp(self) -> None:
        """Set up the test case with a fixed random seed."""
        self.random = np.random.RandomState(seed=42)

    def test_lazy_compile_random_batches(self):
        """Test method lazy_compile_random_batches."""
        indices = np.arange(self.num_samples)
        input_array = self.random.randint(low=0, high=self.num_entities, size=(self.num_samples, 2), dtype=np.int64)
        targets = []
        for _ in range(self.num_samples):
            targets.append(list(set(self.random.randint(low=0, high=self.num_entities, size=(5,), dtype=np.int64))))
        target_array = np.asarray(targets)

        def _batch_compiler(batch_indices):
            return input_array[batch_indices], target_array[batch_indices]

        iterator = lazy_compile_random_batches(
            indices=indices,
            batch_size=self.batch_size,
            batch_compiler=_batch_compiler,
        )
        all_elements = list(iterator)
        for input_batch, target_batch in all_elements[:-1]:
            self.assertEqual(input_batch.shape, (self.batch_size, 2))
            self.assertEqual(target_batch.shape, (self.batch_size,))
        last_input_batch, last_target_batch = all_elements[-1]
        self.assertEqual(last_input_batch.shape, (self.num_samples % self.batch_size, 2))
        self.assertEqual(last_target_batch.shape, (self.num_samples % self.batch_size,))



# -*- coding: utf-8 -*-

"""PyKEEN internal "nn" module."""

from . import init
from .emb import Embedding, EmbeddingSpecification, RepresentationModule

__all__ = [
    'Embedding',
    'EmbeddingSpecification',
    'RepresentationModule',
    'init',
]



# -*- coding: utf-8 -*-

"""Negative sampling algorithm based on the work of of Bordes *et al.*."""

import math
from typing import Collection, Optional

import torch

from .negative_sampler import NegativeSampler

__all__ = [
    'BasicNegativeSampler',
]

LOOKUP = {'h': 0, 'r': 1, 't': 2}


class BasicNegativeSampler(NegativeSampler):
    r"""A basic negative sampler.

    This negative sampler that corrupts positive triples $(h,r,t) \in \mathcal{K}$ by replacing either $h$, $r$ or $t$
    based on the chosen corruption scheme. The corruption scheme can contain $h$, $r$ and $t$ or any subset of these.

    Steps:

    1. Randomly (uniformly) determine whether $h$, $r$ or $t$ shall be corrupted for a positive triple
       $(h,r,t) \in \mathcal{K}$.
    2. Randomly (uniformly) sample an entity $e \in \mathcal{E}$ or relation $r' \in \mathcal{R}$ for selection to
       corrupt the triple.

       - If $h$ was selected before, the corrupted triple is $(e,r,t)$
       - If $r$ was selected before, the corrupted triple is $(h,r',t)$
       - If $t$ was selected before, the corrupted triple is $(h,r,e)$
    3. If ``filtered`` is set to ``True``, all proposed corrupted triples that also exist as
       actual positive triples $(h,r,t) \in \mathcal{K}$ will be removed.
    """

    def __init__(
        self,
        *,
        corruption_scheme: Optional[Collection[str]] = None,
        **kwargs,
    ) -> None:
        """Initialize the basic negative sampler with the given entities.

        :param corruption_scheme:
            What sides ('h', 'r', 't') should be corrupted. Defaults to head and tail ('h', 't').
        :param kwargs:
            Additional keyword based arguments passed to :class:`pykeen.sampling.NegativeSampler`.
        """
        super().__init__(**kwargs)
        self.corruption_scheme = corruption_scheme or ('h', 't')
        # Set the indices
        self._corruption_indices = [LOOKUP[side] for side in self.corruption_scheme]

    def corrupt_batch(self, positive_batch: torch.LongTensor) -> torch.LongTensor:  # noqa: D102
        if self.num_negs_per_pos > 1:
            positive_batch = positive_batch.repeat_interleave(repeats=self.num_negs_per_pos, dim=0)

        # Bind number of negatives to sample
        num_negs = positive_batch.shape[0]

        # Equally corrupt all sides
        split_idx = int(math.ceil(num_negs / len(self._corruption_indices)))

        # Copy positive batch for corruption.
        # Do not detach, as no gradients should flow into the indices.
        negative_batch = positive_batch.clone()

        for index, start in zip(self._corruption_indices, range(0, num_negs, split_idx)):
            stop = min(start + split_idx, num_negs)

            # Relations have a different index maximum than entities
            # At least make sure to not replace the triples by the original value
            index_max = (self.num_relations if index == 1 else self.num_entities) - 1

            negative_batch[start:stop, index] = torch.randint(
                high=index_max,
                size=(stop - start,),
                device=positive_batch.device,
            )

            # To make sure we don't replace the {head, relation, tail} by the
            # original value we shift all values greater or equal than the original value by one up
            # for that reason we choose the random value from [0, num_{heads, relations, tails} -1]
            negative_batch[start:stop, index] += (
                negative_batch[start:stop, index] >= positive_batch[start:stop, index]
            ).long()

        return negative_batch.view(-1, self.num_negs_per_pos, 3)



# -*- coding: utf-8 -*-

"""Plotting utilities for the pipeline results."""

import logging
from typing import Mapping, Optional, Set

from ..losses import loss_resolver
from ..nn.emb import Embedding
from ..stoppers import EarlyStopper

__all__ = [
    'plot_losses',
    'plot_early_stopping',
    'plot_er',
    'plot',
]

logger = logging.getLogger(__name__)

REDUCER_RELATION_WHITELIST = {'PCA'}


def plot_losses(pipeline_result, *, ax=None):
    """Plot the losses per epoch."""
    import seaborn as sns
    sns.set_style('darkgrid')

    ax = _ensure_ax(ax)
    rv = sns.lineplot(x=range(len(pipeline_result.losses)), y=pipeline_result.losses, ax=ax)

    loss_name = loss_resolver.normalize_inst(pipeline_result.model.loss)
    ax.set_ylabel(f'{loss_name} Loss')
    ax.set_xlabel('Epoch')
    ax.set_title(pipeline_result.title if pipeline_result.title is not None else 'Losses Plot')
    return rv


def plot_early_stopping(pipeline_result, *, ax=None, lineplot_kwargs=None):
    """Plot the evaluations during early stopping."""
    import seaborn as sns

    if not isinstance(pipeline_result.stopper, EarlyStopper) or not pipeline_result.stopper.results:
        raise ValueError

    ax = _ensure_ax(ax)

    x = [
        (1 + e) * pipeline_result.stopper.frequency
        for e in range(len(pipeline_result.stopper.results))
    ]
    rv = sns.lineplot(x=x, y=pipeline_result.stopper.results, ax=ax, marker='o', **(lineplot_kwargs or {}))

    ax.set_ylabel(pipeline_result.stopper.metric)
    ax.set_xlabel('Epoch')
    ax.set_title(pipeline_result.title if pipeline_result.title is not None else 'Early Stopper Evaluation Plot')
    return rv


def _default_entity_embedding_getter(m) -> Embedding:
    return m.entity_embeddings


def _default_relation_embedding_getter(m) -> Embedding:
    return m.relation_embeddings


def plot_er(  # noqa: C901
    pipeline_result,
    *,
    model: Optional[str] = None,
    entities: Optional[Set[str]] = None,
    relations: Optional[Set[str]] = None,
    apply_limits: bool = True,
    margin: float = 0.4,
    plot_entities: bool = True,
    plot_relations: Optional[bool] = None,
    annotation_x_offset: float = 0.02,
    annotation_y_offset: float = 0.03,
    entity_embedding_getter=None,
    relation_embedding_getter=None,
    ax=None,
    **kwargs,
):
    """Plot the reduced entities and relation vectors in 2D.

    :param pipeline_result: The result returned by :func:`pykeen.pipeline.pipeline`.
    :param model: The dimensionality reduction model from :mod:`sklearn`. Defaults to PCA.
        Can also use KPCA, GRP, SRP, TSNE, LLE, ISOMAP, MDS, or SE.
    :param entities: A subset of entities to plot
    :param relations: A subset of relations to plot
    :param apply_limits: Should the x and y limits be applied?
    :param margin: The margin size around the minimum/maximum x and y values
    :param plot_entities: If true, plot the entities based on their reduced embeddings
    :param plot_relations: By default, this is only enabled on translational distance models
        like :class:`pykeen.models.TransE`.
    :param annotation_x_offset: X offset of label from entity position
    :param annotation_y_offset: Y offset of label from entity position
    :param entity_embedding_getter: A function that takes a model and returns its entity embeddings. If none,
        defaults to :func:`_default_entity_embedding_getter`, which just gets ``model.entity_embeddings``. Note,
        the default only works with old-style PyKEEN models.
    :param relation_embedding_getter: A function that takes a model and returns its relation embeddings. If none,
        defaults to :func:`_default_relation_embedding_getter`, which just gets ``model.relation_embeddings``. Note,
        the default only works with old-style PyKEEN models.
    :param ax: The matplotlib axis, if pre-defined
    :param kwargs: The keyword arguments passed to `__init__()` of
        the reducer class (e.g., PCA, TSNE)
    :returns: The axis

    :raises ValueError: if entity plotting and relation plotting are both turned off

    .. warning::

        Plotting relations and entities on the same plot is only
        meaningful for translational distance models like TransE.
    """
    import seaborn as sns

    if not plot_entities and not plot_relations:
        raise ValueError

    if plot_relations is None:  # automatically set to true for translational models, false otherwise
        plot_relations = pipeline_result.model.__class__.__name__.lower().startswith('trans')

    if model is None:
        model = 'PCA'
    reducer_cls, reducer_kwargs = _get_reducer_cls(model, **kwargs)
    if plot_relations and reducer_cls.__name__ not in REDUCER_RELATION_WHITELIST:
        raise ValueError(f'Can not use reducer {reducer_cls} when projecting relations. Will result in nonsense')
    reducer = reducer_cls(n_components=2, **reducer_kwargs)

    ax = _ensure_ax(ax)

    sns.set_style('whitegrid')

    if entity_embedding_getter is None:
        entity_embedding_getter = _default_entity_embedding_getter
    if relation_embedding_getter is None:
        relation_embedding_getter = _default_relation_embedding_getter

    if plot_relations and plot_entities:
        e_embeddings, e_reduced = _reduce_embeddings(entity_embedding_getter(pipeline_result.model), reducer, fit=True)
        r_embeddings, r_reduced = _reduce_embeddings(
            relation_embedding_getter(pipeline_result.model), reducer, fit=False,
        )

        xmax = max(r_embeddings[:, 0].max(), e_embeddings[:, 0].max()) + margin
        xmin = min(r_embeddings[:, 0].min(), e_embeddings[:, 0].min()) - margin
        ymax = max(r_embeddings[:, 1].max(), e_embeddings[:, 1].max()) + margin
        ymin = min(r_embeddings[:, 1].min(), e_embeddings[:, 1].min()) - margin
    elif plot_relations:
        e_embeddings, e_reduced = None, False
        r_embeddings, r_reduced = _reduce_embeddings(
            relation_embedding_getter(pipeline_result.model), reducer, fit=True,
        )

        xmax = r_embeddings[:, 0].max() + margin
        xmin = r_embeddings[:, 0].min() - margin
        ymax = r_embeddings[:, 1].max() + margin
        ymin = r_embeddings[:, 1].min() - margin
    elif plot_entities:
        e_embeddings, e_reduced = _reduce_embeddings(entity_embedding_getter(pipeline_result.model), reducer, fit=True)
        r_embeddings, r_reduced = None, False

        xmax = e_embeddings[:, 0].max() + margin
        xmin = e_embeddings[:, 0].min() - margin
        ymax = e_embeddings[:, 1].max() + margin
        ymin = e_embeddings[:, 1].min() - margin
    else:
        raise ValueError  # not even possible

    if not e_reduced and not r_reduced:
        subtitle = ''
    elif reducer_kwargs:
        subtitle = ", ".join("=".join(item) for item in reducer_kwargs.items())
        subtitle = f' using {reducer_cls.__name__} ({subtitle})'
    else:
        subtitle = f' using {reducer_cls.__name__}'

    if plot_entities:
        entity_id_to_label = pipeline_result.training.entity_id_to_label
        for entity_id, entity_reduced_embedding in enumerate(e_embeddings):
            entity_label = entity_id_to_label[entity_id]
            if entities and entity_label not in entities:
                continue
            x, y = entity_reduced_embedding
            ax.scatter(x, y, color='black')
            ax.annotate(entity_label, (x + annotation_x_offset, y + annotation_y_offset))

    if plot_relations:
        relation_id_to_label = pipeline_result.training.relation_id_to_label
        for relation_id, relation_reduced_embedding in enumerate(r_embeddings):
            relation_label = relation_id_to_label[relation_id]
            if relations and relation_label not in relations:
                continue
            x, y = relation_reduced_embedding
            ax.arrow(0, 0, x, y, color='black')
            ax.annotate(relation_label, (x + annotation_x_offset, y + annotation_y_offset))

    if plot_entities and plot_relations:
        ax.set_title(f'Entity/Relation Plot{subtitle}')
    elif plot_entities:
        ax.set_title(f'Entity Plot{subtitle}')
    elif plot_relations:
        ax.set_title(f'Relation Plot{subtitle}')

    if apply_limits:
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])

    return ax


def _ensure_ax(ax):
    if ax is not None:
        return ax

    import matplotlib.pyplot as plt
    return plt.gca()


def _reduce_embeddings(embedding: Embedding, reducer, fit: bool = False):
    embeddings_numpy = embedding(indices=None).detach().cpu().numpy()
    if embeddings_numpy.shape[1] == 2:
        logger.debug('not reducing entity embeddings, already dim=2')
        return embeddings_numpy, False
    elif fit:
        return reducer.fit_transform(embeddings_numpy), True
    else:
        return reducer.transform(embeddings_numpy), True


def _get_reducer_cls(model: str, **kwargs):
    """Get the model class by name and default kwargs.

    :param model: The name of the model. Can choose from: PCA, KPCA, GRP,
        SRP, TSNE, LLE, ISOMAP, MDS, or SE.
    :param kwargs: Keyword arguments that will get passed through and modified based on the chosen model.
    :return: A pair of a reducer class from :mod:`sklearn` and the modified kwargs.

    :raises ValueError: if invalid model name is passed
    """
    if model.upper() == 'PCA':
        from sklearn.decomposition import PCA as Reducer  # noqa:N811
    elif model.upper() == 'KPCA':
        kwargs.setdefault('kernel', 'rbf')
        from sklearn.decomposition import KernelPCA as Reducer
    elif model.upper() == 'GRP':
        from sklearn.random_projection import GaussianRandomProjection as Reducer
    elif model.upper() == 'SRP':
        from sklearn.random_projection import SparseRandomProjection as Reducer
    elif model.upper() in {'T-SNE', 'TSNE'}:
        from sklearn.manifold import TSNE as Reducer  # noqa:N811
    elif model.upper() in {'LLE', 'LOCALLYLINEAREMBEDDING'}:
        from sklearn.manifold import LocallyLinearEmbedding as Reducer
    elif model.upper() == 'ISOMAP':
        from sklearn.manifold import Isomap as Reducer
    elif model.upper() in {'MDS', 'MULTIDIMENSIONALSCALING'}:
        from sklearn.manifold import MDS as Reducer  # noqa:N811
    elif model.upper() in {'SE', 'SPECTRAL', 'SPECTRALEMBEDDING'}:
        from sklearn.manifold import SpectralEmbedding as Reducer
    else:
        raise ValueError(f'invalid dimensionality reduction model: {model}')
    return Reducer, kwargs


def plot(pipeline_result, er_kwargs: Optional[Mapping[str, str]] = None, figsize=(10, 4)):
    """Plot all plots."""
    import matplotlib.pyplot as plt

    fig, axes = plt.subplots(1, 2, figsize=figsize)

    pipeline_result.plot_losses(ax=axes[0])
    pipeline_result.plot_er(ax=axes[1], **(er_kwargs or {}))

    plt.tight_layout()
    return fig, axes



# -*- coding: utf-8 -*-

"""Utilities for generating triples."""

import torch

from .triples_factory import CoreTriplesFactory
from .utils import get_entities, get_relations
from ..typing import TorchRandomHint
from ..utils import ensure_torch_random_state

__all__ = [
    'generate_triples',
    'generate_triples_factory',
]


def generate_triples(
    num_entities: int = 33,
    num_relations: int = 7,
    num_triples: int = 101,
    compact: bool = True,
    random_state: TorchRandomHint = None,
) -> torch.LongTensor:
    """Generate random triples in a torch tensor."""
    random_state = ensure_torch_random_state(random_state)

    rv = torch.stack([
        torch.randint(num_entities, size=(num_triples,), generator=random_state),
        torch.randint(num_relations, size=(num_triples,), generator=random_state),
        torch.randint(num_entities, size=(num_triples,), generator=random_state),
    ], dim=1)

    if compact:
        new_entity_id = {
            entity: i
            for i, entity in enumerate(sorted(get_entities(rv)))
        }
        new_relation_id = {
            relation: i
            for i, relation in enumerate(sorted(get_relations(rv)))
        }
        rv = torch.as_tensor(
            data=[
                [new_entity_id[h], new_relation_id[r], new_entity_id[t]]
                for h, r, t in rv.tolist()
            ],
            dtype=torch.long,
        )

    return rv


def generate_triples_factory(
    num_entities: int = 33,
    num_relations: int = 7,
    num_triples: int = 101,
    random_state: TorchRandomHint = None,
    create_inverse_triples: bool = False,
) -> CoreTriplesFactory:
    """Generate a triples factory with random triples."""
    mapped_triples = generate_triples(
        num_entities=num_entities,
        num_relations=num_relations,
        num_triples=num_triples,
        random_state=random_state,
    )
    return CoreTriplesFactory.create(
        mapped_triples=mapped_triples,
        create_inverse_triples=create_inverse_triples,
    )



# -*- coding: utf-8 -*-

"""Training KGE models based on the sLCWA."""

import logging
from typing import Any, Mapping, Optional

import torch
from class_resolver import HintOrType
from torch.optim.optimizer import Optimizer

from .training_loop import TrainingLoop
from ..losses import CrossEntropyLoss
from ..models import Model
from ..sampling import NegativeSampler, negative_sampler_resolver
from ..triples import CoreTriplesFactory, Instances
from ..triples.instances import SLCWABatchType, SLCWASampleType
from ..typing import MappedTriples

__all__ = [
    'SLCWATrainingLoop',
]

logger = logging.getLogger(__name__)


class SLCWATrainingLoop(TrainingLoop[SLCWASampleType, SLCWABatchType]):
    """A training loop that uses the stochastic local closed world assumption training approach."""

    negative_sampler: NegativeSampler
    loss_blacklist = [CrossEntropyLoss]

    def __init__(
        self,
        model: Model,
        triples_factory: CoreTriplesFactory,
        optimizer: Optional[Optimizer] = None,
        negative_sampler: HintOrType[NegativeSampler] = None,
        negative_sampler_kwargs: Optional[Mapping[str, Any]] = None,
        automatic_memory_optimization: bool = True,
    ):
        """Initialize the training loop.

        :param model: The model to train
        :param triples_factory: The triples factory to train over
        :param optimizer: The optimizer to use while training the model
        :param negative_sampler: The class, instance, or name of the negative sampler
        :param negative_sampler_kwargs: Keyword arguments to pass to the negative sampler class on instantiation
            for every positive one
        :param automatic_memory_optimization:
            Whether to automatically optimize the sub-batch size during
            training and batch size during evaluation with regards to the hardware at hand.
        """
        super().__init__(
            model=model,
            triples_factory=triples_factory,
            optimizer=optimizer,
            automatic_memory_optimization=automatic_memory_optimization,
        )
        self.negative_sampler = negative_sampler_resolver.make(
            query=negative_sampler,
            pos_kwargs=negative_sampler_kwargs,
            triples_factory=triples_factory,
        )

    def _create_instances(self, triples_factory: CoreTriplesFactory) -> Instances:  # noqa: D102
        return triples_factory.create_slcwa_instances()

    @staticmethod
    def _get_batch_size(batch: MappedTriples) -> int:  # noqa: D102
        return batch.shape[0]

    def _process_batch(
        self,
        batch: MappedTriples,
        start: int,
        stop: int,
        label_smoothing: float = 0.0,
        slice_size: Optional[int] = None,
    ) -> torch.FloatTensor:  # noqa: D102
        # Slicing is not possible in sLCWA training loops
        if slice_size is not None:
            raise AttributeError('Slicing is not possible for sLCWA training loops.')

        # Send positive batch to device
        positive_batch = batch[start:stop].to(device=self.device)

        # Create negative samples, shape: (batch_size, num_neg_per_pos, 3)
        negative_batch, positive_filter = self.negative_sampler.sample(positive_batch=positive_batch)

        # apply filter mask
        if positive_filter is None:
            negative_batch = negative_batch.view(-1, 3)
        else:
            negative_batch = negative_batch[positive_filter]

        # Ensure they reside on the device (should hold already for most simple negative samplers, e.g.
        # BasicNegativeSampler, BernoulliNegativeSampler
        negative_batch = negative_batch.to(self.device)

        # Compute negative and positive scores
        positive_scores = self.model.score_hrt(positive_batch)
        negative_scores = self.model.score_hrt(negative_batch).view(*negative_batch.shape[:-1])

        return self.loss.process_slcwa_scores(
            positive_scores=positive_scores,
            negative_scores=negative_scores,
            label_smoothing=label_smoothing,
            batch_filter=positive_filter,
            num_entities=self.model.num_entities,
        ) + self.model.collect_regularization_term()

    def _slice_size_search(
        self,
        *,
        triples_factory: CoreTriplesFactory,
        training_instances: Instances,
        batch_size: int,
        sub_batch_size: int,
        supports_sub_batching: bool,
    ):  # noqa: D102
        # Slicing is not possible for sLCWA
        if supports_sub_batching:
            report = "This model supports sub-batching, but it also requires slicing, which is not possible for sLCWA"
        else:
            report = "This model doesn't support sub-batching and slicing is not possible for sLCWA"
        logger.warning(report)
        raise MemoryError("The current model can't be trained on this hardware with these parameters.")



# -*- coding: utf-8 -*-

"""Inverse Stability Workflow.

This experiment investigates the differences between

"""

import itertools as itt
import logging
from typing import Type

import click
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

import pykeen.evaluation.evaluator
from pykeen.constants import PYKEEN_EXPERIMENTS
from pykeen.datasets import Dataset, get_dataset
from pykeen.models import Model, model_resolver
from pykeen.pipeline import pipeline

INVERSE_STABILITY = PYKEEN_EXPERIMENTS / 'inverse_stability'
INVERSE_STABILITY.mkdir(parents=True, exist_ok=True)

pykeen.evaluation.evaluator.logger.setLevel(logging.CRITICAL)


@click.command()
@click.option('--force', is_flag=True)
@click.option('--clip', type=int, default=10)
def main(force: bool, clip: int):
    """Run the inverse stability experiments."""
    results_path = INVERSE_STABILITY / 'results.tsv'
    if results_path.exists() and not force:
        df = pd.read_csv(results_path, sep='\t')
        df['residuals'] = df['forward'] - df['inverse']
        df = df[(-clip < df['residuals']) & (df['residuals'] < clip)]
        g = sns.FacetGrid(df, col='model', row='dataset', hue='training_loop', sharex=False, sharey=False)
        g.map_dataframe(sns.histplot, x='residuals', stat="density")
        g.add_legend()
        g.savefig(INVERSE_STABILITY / 'results_residuals.png', dpi=300)

    else:
        outer_dfs = []
        datasets = ['nations', 'kinships']
        models = ['rotate', 'complex', 'simple', 'transe', 'distmult']
        training_loops = ['lcwa', 'slcwa']
        for dataset, model, training_loop in itt.product(datasets, models, training_loops):
            click.secho(f'{dataset} {model} {training_loop}', fg='cyan')
            df = run_inverse_stability_workflow(dataset=dataset, model=model, training_loop=training_loop)
            outer_dfs.append(df)
        outer_df = pd.concat(outer_dfs)
        outer_df.to_csv(INVERSE_STABILITY / 'results.tsv', sep='\t', index=False)


def run_inverse_stability_workflow(dataset: str, model: str, training_loop: str, random_seed=0, device='cpu'):
    """Run an inverse stability experiment."""
    dataset_instance: Dataset = get_dataset(
        dataset=dataset,
        dataset_kwargs=dict(
            create_inverse_triples=True,
        ),
    )
    dataset_name = dataset_instance.get_normalized_name()
    model_cls: Type[Model] = model_resolver.lookup(model)
    model_name = model_cls.__name__.lower()

    dataset_dir = INVERSE_STABILITY / dataset_name
    dataset_dir.mkdir(exist_ok=True, parents=True)

    pipeline_result = pipeline(
        dataset=dataset_instance,
        model=model,
        training_loop=training_loop,
        training_kwargs=dict(
            num_epochs=1000,
            use_tqdm_batch=False,
        ),
        stopper='early',
        stopper_kwargs=dict(patience=5, frequency=5),
        random_seed=random_seed,
        device=device,
    )
    test_tf = dataset_instance.testing
    model = pipeline_result.model
    # Score with original triples
    scores_forward = model.score_hrt(test_tf.mapped_triples)
    scores_forward_np = scores_forward.detach().numpy()[:, 0]

    # Score with inverse triples
    scores_inverse = model.score_hrt_inverse(test_tf.mapped_triples)
    scores_inverse_np = scores_inverse.detach().numpy()[:, 0]

    scores_path = dataset_dir / f'{model_name}_{training_loop}_scores.tsv'
    df = pd.DataFrame(
        list(zip(
            itt.repeat(training_loop),
            itt.repeat(dataset_name),
            itt.repeat(model_name),
            scores_forward_np,
            scores_inverse_np,
        )),
        columns=['training_loop', 'dataset', 'model', 'forward', 'inverse'],
    )
    df.to_csv(scores_path, sep='\t', index=False)

    fig, ax = plt.subplots(1, 1)
    sns.histplot(data=df, x='forward', label='Forward', ax=ax, color='blue', stat="density")
    sns.histplot(data=df, x='inverse', label='Inverse', ax=ax, color='orange', stat="density")
    ax.set_title(f'{dataset_name} - {model_name} - {training_loop}')
    ax.set_xlabel('Score')
    plt.legend()
    plt.savefig(dataset_dir / f'{model_name}_{training_loop}_overlay.png', dpi=300)
    plt.close(fig)

    fig, ax = plt.subplots(1, 1)
    sns.histplot(scores_forward_np - scores_inverse_np, ax=ax, stat="density")
    ax.set_title(f'{dataset_name} - {model_name} - {training_loop}')
    ax.set_xlabel('Forward - Inverse Score Difference')
    plt.savefig(dataset_dir / f'{model_name}_{training_loop}_residuals.png', dpi=300)
    plt.close(fig)

    return df


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""The DBpedia datasets from [shi2017b]_.

- GitHub Repository: https://github.com/bxshi/ConMask
- Paper: https://arxiv.org/abs/1711.03438
"""

from docdata import parse_docdata

from .base import UnpackedRemoteDataset

__all__ = [
    'DBpedia50',
]

BASE = 'https://raw.githubusercontent.com/ZhenfengLei/KGDatasets/master/DBpedia50'
TEST_URL = f'{BASE}/test.txt'
TRAIN_URL = f'{BASE}/train.txt'
VALID_URL = f'{BASE}/valid.txt'


@parse_docdata
class DBpedia50(UnpackedRemoteDataset):
    """The DBpedia50 dataset.

    ---
    name: DBpedia50
    citation:
        author: Shi
        year: 2017
        link: https://arxiv.org/abs/1711.03438
    statistics:
        entities: 24624
        relations: 351
        training: 32203
        testing: 2095
        validation: 123
        triples: 34421
    """

    def __init__(self, create_inverse_triples: bool = False, **kwargs):
        """Initialize the DBpedia50 small dataset from [shi2017b]_.

        :param create_inverse_triples: Should inverse triples be created? Defaults to false.
        :param kwargs: keyword arguments passed to :class:`pykeen.datasets.base.UnpackedRemoteDataset`.
        """
        super().__init__(
            training_url=TRAIN_URL,
            testing_url=TEST_URL,
            validation_url=VALID_URL,
            create_inverse_triples=create_inverse_triples,
            load_triples_kwargs={
                # as pointed out in https://github.com/pykeen/pykeen/issues/275#issuecomment-776412294,
                # the columns are not ordered properly.
                'column_remapping': [0, 2, 1],
            },
            **kwargs,
        )


if __name__ == '__main__':
    DBpedia50.cli()



# -*- coding: utf-8 -*-

"""Implementation of wrapper around sklearn metrics."""

from dataclasses import dataclass, field, fields
from typing import Any, Dict, Optional, Tuple

import numpy as np
import torch
from dataclasses_json import dataclass_json
from sklearn import metrics

from .evaluator import Evaluator, MetricResults
from ..typing import MappedTriples
from ..utils import fix_dataclass_init_docs

__all__ = [
    'SklearnEvaluator',
    'SklearnMetricResults',
]


@fix_dataclass_init_docs
@dataclass_json
@dataclass
class SklearnMetricResults(MetricResults):
    """Results from computing metrics."""

    #: The area under the ROC curve
    roc_auc_score: float = field(metadata=dict(
        name="AUC-ROC",
        doc='The area under the ROC curve, on [0, 1]. Higher is better.',
        f=metrics.roc_auc_score,
    ))
    #: The area under the precision-recall curve
    average_precision_score: float = field(metadata=dict(
        name="Average Precision",
        doc='The area under the precision-recall curve, on [0, 1]. Higher is better.',
        f=metrics.average_precision_score,
    ))

    #: The coverage error
    # coverage_error: float = field(metadata=dict(
    #     doc='The coverage error',
    #     f=metrics.coverage_error,
    # ))
    #: The label ranking loss (APS)
    # label_ranking_average_precision_score: float = field(metadata=dict(
    #     doc='The label ranking loss (APS)',
    #     f=metrics.label_ranking_average_precision_score,
    # ))
    # #: The label ranking loss
    # label_ranking_loss: float = field(metadata=dict(
    #     doc='The label ranking loss',
    #     f=metrics.label_ranking_loss,
    # ))

    @classmethod
    def from_scores(cls, y_true, y_score):
        """Return an instance of these metrics from a given set of true and scores."""
        return SklearnMetricResults(**{
            f.name: f.metadata['f'](y_true, y_score)
            for f in fields(cls)
        })

    def get_metric(self, name: str) -> float:  # noqa: D102
        return getattr(self, name)


class SklearnEvaluator(Evaluator):
    """An evaluator that uses a Scikit-learn metric."""

    all_scores: Dict[Tuple[Any, ...], np.ndarray]
    all_positives: Dict[Tuple[Any, ...], np.ndarray]

    def __init__(self, **kwargs):
        super().__init__(
            filtered=False,
            requires_positive_mask=True,
            **kwargs,
        )
        self.all_scores = {}
        self.all_positives = {}

    def _process_scores(
        self,
        keys: torch.LongTensor,
        scores: torch.FloatTensor,
        positive_mask: torch.FloatTensor,
        head_side: bool,
    ) -> None:
        # Transfer to cpu and convert to numpy
        scores = scores.detach().cpu().numpy()
        positive_mask = positive_mask.detach().cpu().numpy()
        keys = keys.detach().cpu().numpy()

        # Ensure that each key gets counted only once
        for i in range(keys.shape[0]):
            # include head_side flag into key to differentiate between (h, r) and (r, t)
            key = (head_side,) + tuple(map(int, keys[i]))
            self.all_scores[key] = scores[i]
            self.all_positives[key] = positive_mask[i]

    def process_tail_scores_(
        self,
        hrt_batch: MappedTriples,
        true_scores: torch.FloatTensor,
        scores: torch.FloatTensor,
        dense_positive_mask: Optional[torch.FloatTensor] = None,
    ) -> None:  # noqa: D102
        if dense_positive_mask is None:
            raise KeyError('Sklearn evaluators need the positive mask!')

        self._process_scores(keys=hrt_batch[:, :2], scores=scores, positive_mask=dense_positive_mask, head_side=False)

    def process_head_scores_(
        self,
        hrt_batch: MappedTriples,
        true_scores: torch.FloatTensor,
        scores: torch.FloatTensor,
        dense_positive_mask: Optional[torch.FloatTensor] = None,
    ) -> None:  # noqa: D102
        if dense_positive_mask is None:
            raise KeyError('Sklearn evaluators need the positive mask!')

        self._process_scores(keys=hrt_batch[:, 1:], scores=scores, positive_mask=dense_positive_mask, head_side=True)

    def finalize(self) -> SklearnMetricResults:  # noqa: D102
        # Important: The order of the values of an dictionary is not guaranteed. Hence, we need to retrieve scores and
        # masks using the exact same key order.
        all_keys = list(self.all_scores.keys())
        # TODO how to define a cutoff on y_scores to make binary?
        # see: https://github.com/xptree/NetMF/blob/77286b826c4af149055237cef65e2a500e15631a/predict.py#L25-L33
        y_score = np.concatenate([self.all_scores[k] for k in all_keys], axis=0).flatten()
        y_true = np.concatenate([self.all_positives[k] for k in all_keys], axis=0).flatten()

        # Clear buffers
        self.all_positives.clear()
        self.all_scores.clear()

        return SklearnMetricResults.from_scores(y_true, y_score)



# -*- coding: utf-8 -*-

"""Test the PyKEEN pipeline function."""

import tempfile
import unittest

import pandas as pd
import torch

import pykeen.regularizers
from pykeen.datasets import EagerDataset, Nations
from pykeen.models import ERModel, Model
from pykeen.models.predict import (
    get_all_prediction_df, get_head_prediction_df, get_relation_prediction_df,
    get_tail_prediction_df,
)
from pykeen.models.resolve import DimensionError, make_model, make_model_cls
from pykeen.nn.modules import TransEInteraction
from pykeen.pipeline import PipelineResult, pipeline
from pykeen.regularizers import NoRegularizer
from pykeen.training import SLCWATrainingLoop
from pykeen.triples.generation import generate_triples_factory
from pykeen.utils import resolve_device
from tests.mocks import MockModel


class TestPipeline(unittest.TestCase):
    """Test the pipeline."""

    @classmethod
    def setUpClass(cls):
        """Set up a shared result."""
        cls.device = resolve_device('cuda')
        cls.dataset = Nations()
        cls.result = pipeline(
            model='TransE',
            dataset=cls.dataset,
            training_kwargs=dict(num_epochs=5, use_tqdm=False),
            evaluation_kwargs=dict(use_tqdm=False),
            device=cls.device,
            random_seed=42,
        )
        cls.model = cls.result.model
        cls.testing_mapped_triples = cls.dataset.testing.mapped_triples.to(cls.model.device)

    def test_predict_tails_no_novelties(self):
        """Test scoring tails without labeling as novel w.r.t. training and testing."""
        tails_df = get_tail_prediction_df(
            self.model,
            'brazil', 'intergovorgs', testing=self.testing_mapped_triples,
            triples_factory=self.dataset.training,
            add_novelties=False,
        )
        self.assertEqual(['tail_id', 'tail_label', 'score'], list(tails_df.columns))
        self.assertEqual(len(self.dataset.training.entity_to_id), len(tails_df.index))

    def test_predict_tails_remove_known(self):
        """Test scoring tails while removing non-novel triples w.r.t. training and testing."""
        tails_df = get_tail_prediction_df(
            self.model,
            'brazil', 'intergovorgs', testing=self.testing_mapped_triples,
            remove_known=True,
            triples_factory=self.dataset.training,
        )
        self.assertEqual(['tail_id', 'tail_label', 'score'], list(tails_df.columns))
        self.assertEqual({'jordan', 'brazil', 'ussr', 'burma', 'china'}, set(tails_df['tail_label']))

    def test_predict_tails_with_novelties(self):
        """Test scoring tails with labeling as novel w.r.t. training and testing."""
        tails_df = get_tail_prediction_df(
            self.model, 'brazil', 'intergovorgs',
            triples_factory=self.dataset.training,
            testing=self.testing_mapped_triples,
        )
        self.assertEqual(['tail_id', 'tail_label', 'score', 'in_training', 'in_testing'], list(tails_df.columns))
        self.assertEqual(self.model.num_entities, len(tails_df.index))
        training_tails = set(tails_df.loc[tails_df['in_training'], 'tail_label'])
        self.assertEqual({'usa', 'uk', 'netherlands', 'egypt', 'india', 'israel', 'indonesia'}, training_tails)
        testing_tails = set(tails_df.loc[tails_df['in_testing'], 'tail_label'])
        self.assertEqual({'poland', 'cuba'}, testing_tails)

    def test_predict_relations_with_novelties(self):
        """Test scoring relations with labeling as novel w.r.t. training and testing."""
        rel_df = get_relation_prediction_df(
            self.model, 'brazil', 'uk',
            triples_factory=self.dataset.training,
            testing=self.testing_mapped_triples,
        )
        self.assertEqual(['relation_id', 'relation_label', 'score', 'in_training', 'in_testing'], list(rel_df.columns))
        self.assertEqual(self.model.num_relations, len(rel_df.index))
        training_rels = set(rel_df.loc[rel_df['in_training'], 'relation_label'])
        self.assertEqual(
            {
                'weightedunvote', 'relexports', 'intergovorgs', 'timesinceally', 'exports3', 'booktranslations',
                'relbooktranslations', 'reldiplomacy', 'ngoorgs3', 'ngo', 'relngo', 'reltreaties', 'independence',
                'intergovorgs3', 'unweightedunvote', 'commonbloc2', 'relintergovorgs',
            },
            training_rels,
        )
        testing_heads = set(rel_df.loc[rel_df['in_testing'], 'relation_label'])
        self.assertEqual({'embassy'}, testing_heads)

    def test_predict_heads_with_novelties(self):
        """Test scoring heads with labeling as novel w.r.t. training and testing."""
        heads_df = get_head_prediction_df(
            self.model,
            'conferences', 'brazil',
            triples_factory=self.dataset.training,
            testing=self.testing_mapped_triples,
        )
        self.assertEqual(['head_id', 'head_label', 'score', 'in_training', 'in_testing'], list(heads_df.columns))
        self.assertEqual(self.model.num_entities, len(heads_df.index))
        training_heads = set(heads_df.loc[heads_df['in_training'], 'head_label'])
        self.assertEqual({'usa', 'india', 'ussr', 'poland', 'cuba'}, training_heads)
        testing_heads = set(heads_df.loc[heads_df['in_testing'], 'head_label'])
        self.assertEqual(set(), testing_heads)

    def test_predict_all_no_novelties(self):
        """Test scoring all triples without labeling as novel w.r.t. training and testing."""
        all_df = get_all_prediction_df(
            model=self.model,
            triples_factory=self.dataset.training,
            testing=self.testing_mapped_triples,
            add_novelties=False,
        )
        self.assertIsInstance(all_df, pd.DataFrame)
        self.assertEqual(
            ['head_id', 'head_label', 'relation_id', 'relation_label', 'tail_id', 'tail_label', 'score'],
            list(all_df.columns),
        )
        possible = self.dataset.training.num_relations * self.model.num_entities ** 2
        self.assertEqual(possible, len(all_df.index))

    def test_predict_all_remove_known(self):
        """Test scoring all triples while removing non-novel triples w.r.t. training and testing."""
        all_df = get_all_prediction_df(
            model=self.model,
            triples_factory=self.dataset.training,
            testing=self.testing_mapped_triples,
            remove_known=True,
        )
        self.assertIsInstance(all_df, pd.DataFrame)
        self.assertEqual(
            ['head_id', 'head_label', 'relation_id', 'relation_label', 'tail_id', 'tail_label', 'score'],
            list(all_df.columns),
        )
        possible = self.dataset.training.num_relations * self.model.num_entities ** 2
        known = self.dataset.training.num_triples + self.testing_mapped_triples.shape[0]
        self.assertNotEqual(possible, known, msg='testing and training triples cover all possible triples')
        self.assertEqual(possible - known, len(all_df.index))

    def test_predict_all_with_novelties(self):
        """Test scoring all triples with labeling as novel w.r.t. training and testing."""
        all_df = get_all_prediction_df(
            model=self.model,
            triples_factory=self.dataset.training,
            testing=self.testing_mapped_triples,
        )
        self.assertIsInstance(all_df, pd.DataFrame)
        self.assertEqual(
            [
                'head_id', 'head_label', 'relation_id', 'relation_label',
                'tail_id', 'tail_label', 'score', 'in_training', 'in_testing',
            ],
            list(all_df.columns),
        )
        possible = self.dataset.training.num_relations * self.model.num_entities ** 2
        self.assertEqual(possible, len(all_df.index))
        self.assertEqual(self.dataset.training.num_triples, all_df['in_training'].sum())
        self.assertEqual(self.testing_mapped_triples.shape[0], all_df['in_testing'].sum())


class TestPipelineTriples(unittest.TestCase):
    """Test applying the pipeline to triples factories."""

    def setUp(self) -> None:
        """Prepare the training, testing, and validation triples factories."""
        self.base_tf = generate_triples_factory(
            num_entities=50,
            num_relations=9,
            num_triples=500,
        )
        self.training, self.testing, self.validation = self.base_tf.split([0.8, 0.1, 0.1])

    def test_unlabeled_triples(self):
        """Test running the pipeline on unlabeled triples factories."""
        _ = pipeline(
            training=self.training,
            testing=self.testing,
            validation=self.validation,
            model='TransE',
            training_kwargs=dict(num_epochs=1, use_tqdm=False),
            evaluation_kwargs=dict(use_tqdm=False),
        )

    def test_eager_unlabeled_dataset(self):
        """Test running the pipeline on unlabeled triples factories in a dataset."""
        dataset = EagerDataset(
            training=self.training,
            testing=self.testing,
            validation=self.validation,
        )
        _ = pipeline(
            dataset=dataset,
            model='TransE',
            training_kwargs=dict(num_epochs=1, use_tqdm=False),
            evaluation_kwargs=dict(use_tqdm=False),
        )

    def test_interaction_instance_missing_dimensions(self):
        """Test when a dimension is missing."""
        with self.assertRaises(DimensionError) as exc:
            make_model_cls(
                dimensions={},  # missing "d"
                interaction=TransEInteraction(p=2),
            )
        self.assertIsInstance(exc.exception, DimensionError)
        self.assertEqual({'d'}, exc.exception.expected)
        self.assertEqual(set(), exc.exception.given)
        self.assertEqual("Expected dimensions dictionary with keys {'d'} but got keys set()", str(exc.exception))

    def test_interaction_instance_builder(self):
        """Test resolving an interaction model instance."""
        model = make_model(
            dimensions={"d": 3},
            interaction=TransEInteraction,
            interaction_kwargs=dict(p=2),
            triples_factory=self.training,
        )
        self.assertIsInstance(model, ERModel)
        self.assertIsInstance(model.interaction, TransEInteraction)
        self.assertEqual(2, model.interaction.p)
        _ = pipeline(
            training=self.training,
            testing=self.testing,
            validation=self.validation,
            model=model,
            training_kwargs=dict(num_epochs=1, use_tqdm=False),
            evaluation_kwargs=dict(use_tqdm=False),
            random_seed=0,
        )

    def test_interaction_builder(self):
        """Test resolving an interaction model."""
        model_cls = make_model_cls({"d": 3}, TransEInteraction(p=2))
        self._help_test_interaction_resolver(model_cls)

    def test_interaction_resolver_cls(self):
        """Test resolving the interaction function."""
        model_cls = make_model_cls({"d": 3}, TransEInteraction, {'p': 2})
        self._help_test_interaction_resolver(model_cls)

    def test_interaction_resolver_lookup(self):
        """Test resolving the interaction function."""
        model_cls = make_model_cls({"d": 3}, 'TransE', {'p': 2})
        self._help_test_interaction_resolver(model_cls)

    def _help_test_interaction_resolver(self, model_cls):
        self.assertTrue(issubclass(model_cls, ERModel))
        self.assertIsInstance(model_cls._interaction, TransEInteraction)
        self.assertEqual(2, model_cls._interaction.p)
        _ = pipeline(
            training=self.training,
            testing=self.testing,
            validation=self.validation,
            model=model_cls,
            training_kwargs=dict(num_epochs=1, use_tqdm=False),
            evaluation_kwargs=dict(use_tqdm=False),
            random_seed=0,
        )

    def test_custom_training_loop(self):
        """Test providing a custom training loop."""
        losses = []

        class ModifiedTrainingLoop(SLCWATrainingLoop):
            """A wrapper around SLCWA training loop which remembers batch losses."""

            def _forward_pass(self, *args, **kwargs):  # noqa: D102
                loss = super()._forward_pass(*args, **kwargs)
                losses.append(loss)
                return loss

        _ = pipeline(
            training=self.training,
            testing=self.testing,
            validation=self.validation,
            training_loop=ModifiedTrainingLoop,
            model='TransE',
            training_kwargs=dict(num_epochs=1, use_tqdm=False),
            evaluation_kwargs=dict(use_tqdm=False),
            random_seed=0,
        )

        # empty lists are falsy
        self.assertTrue(losses)


class TestPipelineCheckpoints(unittest.TestCase):
    """Test the pipeline with checkpoints."""

    def setUp(self) -> None:
        """Set up a shared result as standard to compare to."""
        self.random_seed = 123
        self.model = 'TransE'
        self.dataset = 'nations'
        self.checkpoint_name = "PyKEEN_training_loop_test_checkpoint.pt"
        self.temporary_directory = tempfile.TemporaryDirectory()

    def tearDown(self) -> None:
        """Tear down the test case."""
        self.temporary_directory.cleanup()

    def test_pipeline_resumption(self):
        """Test whether the resumed LCWA pipeline creates the same results as the one shot pipeline."""
        self._test_pipeline_x_resumption(training_loop_type='LCWA')

    def test_pipeline_slcwa_resumption(self):
        """Test whether the resumed sLCWA pipeline creates the same results as the one shot pipeline."""
        self._test_pipeline_x_resumption(training_loop_type='sLCWA')

    def _test_pipeline_x_resumption(self, training_loop_type: str):
        """Test whether the resumed pipeline creates the same results as the one shot pipeline."""
        # As the resumption capability currently is a function of the training loop, more thorough tests can be found
        # in the test_training.py unit tests. In the tests below the handling of training loop checkpoints by the
        # pipeline is checked.

        result_standard = pipeline(
            model=self.model,
            dataset=self.dataset,
            training_loop=training_loop_type,
            training_kwargs=dict(num_epochs=10, use_tqdm=False, use_tqdm_batch=False),
            random_seed=self.random_seed,
        )

        # Set up a shared result that runs two pipelines that should replicate the results of the standard pipeline.
        _ = pipeline(
            model=self.model,
            dataset=self.dataset,
            training_loop=training_loop_type,
            training_kwargs=dict(
                num_epochs=5,
                use_tqdm=False,
                use_tqdm_batch=False,
                checkpoint_name=self.checkpoint_name,
                checkpoint_directory=self.temporary_directory.name,
                checkpoint_frequency=0,
            ),
            random_seed=self.random_seed,
        )

        # Resume the previous pipeline
        result_split = pipeline(
            model=self.model,
            dataset=self.dataset,
            training_loop=training_loop_type,
            training_kwargs=dict(
                num_epochs=10,
                use_tqdm=False,
                use_tqdm_batch=False,
                checkpoint_name=self.checkpoint_name,
                checkpoint_directory=self.temporary_directory.name,
                checkpoint_frequency=0,
            ),
        )
        self.assertEqual(result_standard.losses, result_split.losses)


class TestAttributes(unittest.TestCase):
    """Test that the keywords given to the pipeline make it through."""

    def test_specify_regularizer(self):
        """Test a pipeline that uses a regularizer."""
        for regularizer, cls in [
            (None, pykeen.regularizers.NoRegularizer),
            ('no', pykeen.regularizers.NoRegularizer),
            (NoRegularizer, pykeen.regularizers.NoRegularizer),
            ('powersum', pykeen.regularizers.PowerSumRegularizer),
            ('lp', pykeen.regularizers.LpRegularizer),
        ]:
            with self.subTest(regularizer=regularizer):
                pipeline_result = pipeline(
                    model='TransE',
                    dataset='Nations',
                    regularizer=regularizer,
                    training_kwargs=dict(num_epochs=1),
                )
                self.assertIsInstance(pipeline_result, PipelineResult)
                self.assertIsInstance(pipeline_result.model, Model)
                self.assertIsInstance(pipeline_result.model.regularizer, cls)


class TestPipelineEvaluationFiltering(unittest.TestCase):
    """Test filtering of triples during evaluation using the pipeline."""

    @classmethod
    def setUpClass(cls):
        """Set up a shared result."""
        cls.device = resolve_device('cuda')
        cls.dataset = Nations()

        cls.model = MockModel(triples_factory=cls.dataset.training)

        # The MockModel gives the highest score to the highest entity id
        max_score = cls.dataset.num_entities - 1

        # The test triples are created to yield the third highest score on both head and tail prediction
        cls.dataset.testing.mapped_triples = torch.tensor([[max_score - 2, 0, max_score - 2]])

        # Write new mapped triples to the model, since the model's triples will be used to filter
        # These triples are created to yield the highest score on both head and tail prediction for the
        # test triple at hand
        cls.dataset.training.mapped_triples = torch.tensor(
            [
                [max_score - 2, 0, max_score],
                [max_score, 0, max_score - 2],
            ],
        )

        # The validation triples are created to yield the second highest score on both head and tail prediction for the
        # test triple at hand
        cls.dataset.validation.mapped_triples = torch.tensor(
            [
                [max_score - 2, 0, max_score - 1],
                [max_score - 1, 0, max_score - 2],
            ],
        )

    def test_pipeline_evaluation_filtering_without_validation_triples(self):
        """Test if the evaluator's triple filtering works as expected using the pipeline."""
        results = pipeline(
            model=self.model,
            dataset=self.dataset,
            training_loop_kwargs=dict(automatic_memory_optimization=False),
            training_kwargs=dict(num_epochs=0, use_tqdm=False),
            evaluator_kwargs=dict(filtered=True, automatic_memory_optimization=False),
            evaluation_kwargs=dict(use_tqdm=False),
            device=self.device,
            random_seed=42,
            filter_validation_when_testing=False,
        )
        assert results.metric_results.arithmetic_mean_rank['both']['realistic'] == 2, 'The rank should equal 2'

    def test_pipeline_evaluation_filtering_with_validation_triples(self):
        """Test if the evaluator's triple filtering with validation triples works as expected using the pipeline."""
        results = pipeline(
            model=self.model,
            dataset=self.dataset,
            training_loop_kwargs=dict(automatic_memory_optimization=False),
            training_kwargs=dict(num_epochs=0, use_tqdm=False),
            evaluator_kwargs=dict(filtered=True, automatic_memory_optimization=False),
            evaluation_kwargs=dict(use_tqdm=False),
            device=self.device,
            random_seed=42,
            filter_validation_when_testing=True,
        )
        assert results.metric_results.arithmetic_mean_rank['both']['realistic'] == 1, 'The rank should equal 1'



# -*- coding: utf-8 -*-

"""WordNet datasets."""

import pathlib

from docdata import parse_docdata

from .base import TarFileRemoteDataset

__all__ = [
    'WN18',
    'WN18RR',
]


@parse_docdata
class WN18(TarFileRemoteDataset):
    """The WN18 dataset.

    ---
    name: WordNet-18
    statistics:
        entities: 40943
        relations: 18
        training: 141442
        testing: 5000
        validation: 5000
        triples: 151442
    citation:
        author: Bordes
        year: 2014
        link: https://arxiv.org/abs/1301.3485
    """

    def __init__(self, create_inverse_triples: bool = False, **kwargs):
        """Initialize the WordNet-18 dataset.

        :param create_inverse_triples: Should inverse triples be created? Defaults to false.
        :param kwargs: keyword arguments passed to :class:`pykeen.datasets.base.TarFileRemoteDataset`.

        .. warning:: This dataset contains testing leakage. Use :class:`WN18RR` instead.
        """
        super().__init__(
            url='https://everest.hds.utc.fr/lib/exe/fetch.php?media=en:wordnet-mlj12.tar.gz',
            relative_training_path=pathlib.PurePath('wordnet-mlj12', 'wordnet-mlj12-train.txt'),
            relative_testing_path=pathlib.PurePath('wordnet-mlj12', 'wordnet-mlj12-test.txt'),
            relative_validation_path=pathlib.PurePath('wordnet-mlj12', 'wordnet-mlj12-valid.txt'),
            create_inverse_triples=create_inverse_triples,
            **kwargs,
        )


@parse_docdata
class WN18RR(TarFileRemoteDataset):
    """The WN18-RR dataset.

    ---
    name: WordNet-18 (RR)
    statistics:
        entities: 40559
        relations: 11
        training: 86835
        testing: 2924
        validation: 2824
        triples: 92583
    citation:
        author: Toutanova
        year: 2015
        link: https://www.aclweb.org/anthology/W15-4007/
    """

    def __init__(self, create_inverse_triples: bool = False, **kwargs):
        """Initialize the WordNet-18 (RR) dataset.

        :param create_inverse_triples: Should inverse triples be created? Defaults to false.
        :param kwargs: keyword arguments passed to :class:`pykeen.datasets.base.TarFileRemoteDataset`.
        """
        super().__init__(
            url='https://github.com/TimDettmers/ConvE/raw/master/WN18RR.tar.gz',
            relative_training_path='train.txt',
            relative_testing_path='test.txt',
            relative_validation_path='valid.txt',
            create_inverse_triples=create_inverse_triples,
            **kwargs,
        )


def _main():
    for cls in [WN18, WN18RR]:
        cls().summarize()


if __name__ == '__main__':
    _main()



# -*- coding: utf-8 -*-

"""New-style base module for all KGE models."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from operator import itemgetter
from typing import Any, ClassVar, Generic, Iterable, List, Mapping, Optional, Sequence, Tuple, Type, Union, cast

import torch
from torch import nn

from .base import Model
from ..losses import Loss
from ..nn.emb import EmbeddingSpecification, RepresentationModule
from ..nn.modules import Interaction, interaction_resolver
from ..regularizers import Regularizer
from ..triples import CoreTriplesFactory
from ..typing import DeviceHint, HeadRepresentation, RelationRepresentation, TailRepresentation
from ..utils import check_shapes

__all__ = [
    '_NewAbstractModel',
    'ERModel',
]

logger = logging.getLogger(__name__)

EmbeddingSpecificationHint = Union[
    None,
    EmbeddingSpecification,
    RepresentationModule,
    Sequence[Union[EmbeddingSpecification, RepresentationModule]],
]


class _NewAbstractModel(Model, ABC):
    """An abstract class for knowledge graph embedding models (KGEMs).

    The only function that needs to be implemented for a given subclass is
    :meth:`Model.forward`. The job of the :meth:`Model.forward` function, as
    opposed to the completely general :meth:`torch.nn.Module.forward` is
    to take indices for the head, relation, and tails' respective representation(s)
    and to determine a score.

    Subclasses of Model can decide however they want on how to store entities' and
    relations' representations, how they want to be looked up, and how they should
    be scored. The :class:`ERModel` provides a commonly useful implementation
    which allows for the specification of one or more entity representations and
    one or more relation representations in the form of :class:`pykeen.nn.Embedding`
    as well as a matching instance of a :class:`pykeen.nn.Interaction`.
    """

    #: The default regularizer class
    regularizer_default: ClassVar[Optional[Type[Regularizer]]] = None
    #: The default parameters for the default regularizer class
    regularizer_default_kwargs: ClassVar[Optional[Mapping[str, Any]]] = None

    def _reset_parameters_(self):  # noqa: D401
        """Reset all parameters of the model in-place."""
        # cf. https://github.com/mberr/ea-sota-comparison/blob/6debd076f93a329753d819ff4d01567a23053720/src/kgm/utils/torch_utils.py#L317-L372   # noqa:E501
        # Make sure that all modules with parameters do have a reset_parameters method.
        uninitialized_parameters = set(map(id, self.parameters()))
        parents = defaultdict(list)

        # Recursively visit all sub-modules
        task_list = []
        for name, module in self.named_modules():
            # skip self
            if module is self:
                continue

            # Track parents for blaming
            for p in module.parameters():
                parents[id(p)].append(module)

            # call reset_parameters if possible
            if hasattr(module, 'reset_parameters'):
                task_list.append((name.count('.'), module))

        # initialize from bottom to top
        # This ensures that specialized initializations will take priority over the default ones of its components.
        for module in map(itemgetter(1), sorted(task_list, reverse=True, key=itemgetter(0))):
            module.reset_parameters()
            uninitialized_parameters.difference_update(map(id, module.parameters()))

        # emit warning if there where parameters which were not initialised by reset_parameters.
        if len(uninitialized_parameters) > 0:
            logger.warning(
                'reset_parameters() not found for all modules containing parameters. '
                '%d parameters where likely not initialized.',
                len(uninitialized_parameters),
            )

            # Additional debug information
            for i, p_id in enumerate(uninitialized_parameters, start=1):
                logger.debug('[%3d] Parents to blame: %s', i, parents.get(p_id))

    def _instantiate_default_regularizer(self, **kwargs) -> Optional[Regularizer]:
        """Instantiate the regularizer from this class's default settings.

        :param kwargs: Additional keyword arguments to be passed through to the ``__init__()`` function of the
            default regularizer, if one is set.

        :returns: If the default regularizer is None, None is returned.
        """
        if self.regularizer_default is None:
            return None

        _kwargs = dict(self.regularizer_default_kwargs or {})
        _kwargs.update(kwargs)
        return self.regularizer_default(**_kwargs)

    def post_parameter_update(self) -> None:
        """Has to be called after each parameter update."""
        for module in self.modules():
            if module is self:
                continue
            if hasattr(module, "post_parameter_update"):
                module.post_parameter_update()

    def collect_regularization_term(self):  # noqa: D102
        return sum(
            regularizer.pop_regularization_term()
            for regularizer in self.modules()
            if isinstance(regularizer, Regularizer)
        )

    @abstractmethod
    def forward(
        self,
        h_indices: Optional[torch.LongTensor],
        r_indices: Optional[torch.LongTensor],
        t_indices: Optional[torch.LongTensor],
        slice_size: Optional[int] = None,
        slice_dim: Optional[str] = None,
    ) -> torch.FloatTensor:
        """Forward pass.

        This method takes head, relation and tail indices and calculates the corresponding score.

        .. note ::
            All indices which are not None, have to be either 1-element, be of shape `(batch_size,)` or
            `(batch_size, n)`, where `batch_size` has to be the same for all tensors, but `n` may be different.

        .. note ::
            If slicing is requested, the corresponding indices have to be None.

        :param h_indices:
            The head indices. None indicates to use all.
        :param r_indices:
            The relation indices. None indicates to use all.
        :param t_indices:
            The tail indices. None indicates to use all.
        :param slice_size:
            The slice size.
        :param slice_dim:
            The dimension along which to slice. From {"h", "r", "t"}.

        :return: shape: (batch_size, num_heads, num_relations, num_tails)
            The score for each triple.
        """  # noqa: DAR202

    def score_hrt(self, hrt_batch: torch.LongTensor) -> torch.FloatTensor:
        """Forward pass.

        This method takes head, relation and tail of each triple and calculates the corresponding score.

        :param hrt_batch: shape: (batch_size, 3), dtype: long
            The indices of (head, relation, tail) triples.

        :return: shape: (batch_size, 1), dtype: float
            The score for each triple.
        """
        return self(
            h_indices=hrt_batch[:, 0],
            r_indices=hrt_batch[:, 1],
            t_indices=hrt_batch[:, 2],
        ).view(hrt_batch.shape[0], 1)

    def score_t(self, hr_batch: torch.LongTensor, slice_size: Optional[int] = None) -> torch.FloatTensor:
        """Forward pass using right side (tail) prediction.

        This method calculates the score for all possible tails for each (head, relation) pair.

        :param hr_batch: shape: (batch_size, 2), dtype: long
            The indices of (head, relation) pairs.
        :param slice_size:
            The slice size.

        :return: shape: (batch_size, num_entities), dtype: float
            For each h-r pair, the scores for all possible tails.
        """
        return self(
            h_indices=hr_batch[:, 0],
            r_indices=hr_batch[:, 1],
            t_indices=None,
            slice_size=slice_size,
            slice_dim="h",
        ).view(hr_batch.shape[0], self.num_entities)

    def score_h(self, rt_batch: torch.LongTensor, slice_size: Optional[int] = None) -> torch.FloatTensor:
        """Forward pass using left side (head) prediction.

        This method calculates the score for all possible heads for each (relation, tail) pair.

        :param rt_batch: shape: (batch_size, 2), dtype: long
            The indices of (relation, tail) pairs.
        :param slice_size:
            The slice size.

        :return: shape: (batch_size, num_entities), dtype: float
            For each r-t pair, the scores for all possible heads.
        """
        return self(
            h_indices=None,
            r_indices=rt_batch[:, 0],
            t_indices=rt_batch[:, 1],
            slice_size=slice_size,
            slice_dim="r",
        ).view(rt_batch.shape[0], self.num_entities)

    def score_r(self, ht_batch: torch.LongTensor, slice_size: Optional[int] = None) -> torch.FloatTensor:
        """Forward pass using middle (relation) prediction.

        This method calculates the score for all possible relations for each (head, tail) pair.

        :param ht_batch: shape: (batch_size, 2), dtype: long
            The indices of (head, tail) pairs.
        :param slice_size:
            The slice size.

        :return: shape: (batch_size, num_relations), dtype: float
            For each h-t pair, the scores for all possible relations.
        """
        return self(
            h_indices=ht_batch[:, 0],
            r_indices=None,
            t_indices=ht_batch[:, 1],
            slice_size=slice_size,
            slice_dim="t",
        ).view(ht_batch.shape[0], self.num_relations)


def _prepare_representation_module_list(
    representations: EmbeddingSpecificationHint,
    num_embeddings: int,
    shapes: Sequence[str],
    label: str,
    skip_checks: bool = False,
) -> Sequence[RepresentationModule]:
    """Normalize list of representations and wrap into nn.ModuleList."""
    # Important: use ModuleList to ensure that Pytorch correctly handles their devices and parameters
    if representations is None:
        representations = []
    if not isinstance(representations, Sequence):
        representations = [representations]
    if not skip_checks and len(representations) != len(shapes):
        raise ValueError(
            f"Interaction function requires {len(shapes)} {label} representations, but "
            f"{len(representations)} were given.",
        )
    modules = []
    for r in representations:
        if not isinstance(r, RepresentationModule):
            assert isinstance(r, EmbeddingSpecification)
            r = r.make(num_embeddings=num_embeddings)
        if r.max_id < num_embeddings:
            raise ValueError(
                f"{r} only provides {r.max_id} {label} representations, but should provide {num_embeddings}.",
            )
        elif r.max_id > num_embeddings:
            logger.warning(
                f"{r} provides {r.max_id} {label} representations, although only {num_embeddings} are needed."
                f"While this is not necessarily wrong, it can indicate an error where the number of {label} "
                f"representations was chosen wrong.",
            )
        modules.append(r)
    if not skip_checks:
        check_shapes(*zip(
            (r.shape for r in modules),
            shapes,
        ), raise_on_errors=True)
    return nn.ModuleList(modules)


class ERModel(
    Generic[HeadRepresentation, RelationRepresentation, TailRepresentation],
    _NewAbstractModel,
):
    """A commonly useful base for KGEMs using embeddings and interaction modules.

    This model does not use post-init hooks to automatically initialize all of its
    parameters. Rather, the call to :func:`Model.reset_parameters_` happens at the end of
    ``ERModel.__init__``. This is possible because all trainable parameters should necessarily
    be passed through the ``super().__init__()`` in subclasses of :class:`ERModel`.

    Other code can still be put after the call to ``super().__init__()`` in subclasses, such as
    registering regularizers (as done in :class:`pykeen.models.ConvKB` and :class:`pykeen.models.TransH`).
    ---
    citation:
        author: Ali
        year: 2021
        link: https://jmlr.org/papers/v22/20-825.html
        github: pykeen/pykeen
    """

    #: The entity representations
    entity_representations: Sequence[RepresentationModule]

    #: The relation representations
    relation_representations: Sequence[RepresentationModule]

    #: The weight regularizers
    weight_regularizers: List[Regularizer]

    def __init__(
        self,
        *,
        triples_factory: CoreTriplesFactory,
        interaction: Union[
            str,
            Interaction[HeadRepresentation, RelationRepresentation, TailRepresentation],
            Type[Interaction[HeadRepresentation, RelationRepresentation, TailRepresentation]],
        ],
        interaction_kwargs: Optional[Mapping[str, Any]] = None,
        entity_representations: EmbeddingSpecificationHint = None,
        relation_representations: EmbeddingSpecificationHint = None,
        loss: Optional[Loss] = None,
        predict_with_sigmoid: bool = False,
        preferred_device: DeviceHint = None,
        random_seed: Optional[int] = None,
    ) -> None:
        """Initialize the module.

        :param triples_factory:
            The triples factory facilitates access to the dataset.
        :param interaction: The interaction module (e.g., TransE)
        :param interaction_kwargs:
            Additional key-word based parameters given to the interaction module's constructor, if not already
            instantiated.
        :param entity_representations: The entity representation or sequence of representations
        :param relation_representations: The relation representation or sequence of representations
        :param loss:
            The loss to use. If None is given, use the loss default specific to the model subclass.
        :param predict_with_sigmoid:
            Whether to apply sigmoid onto the scores when predicting scores. Applying sigmoid at prediction time may
            lead to exactly equal scores for certain triples with very high, or very low score. When not trained with
            applying sigmoid (or using BCEWithLogitsLoss), the scores are not calibrated to perform well with sigmoid.
        :param preferred_device:
            The preferred device for model training and inference.
        :param random_seed:
            A random seed to use for initialising the model's weights. **Should** be set when aiming at reproducibility.
        """
        super().__init__(
            triples_factory=triples_factory,
            loss=loss,
            preferred_device=preferred_device,
            random_seed=random_seed,
            predict_with_sigmoid=predict_with_sigmoid,
        )
        self.interaction = interaction_resolver.make(interaction, pos_kwargs=interaction_kwargs)
        self.entity_representations = _prepare_representation_module_list(
            representations=entity_representations,
            num_embeddings=triples_factory.num_entities,
            shapes=self.interaction.entity_shape,
            label="entity",
            skip_checks=self.interaction.tail_entity_shape is not None,
        )
        self.relation_representations = _prepare_representation_module_list(
            representations=relation_representations,
            num_embeddings=triples_factory.num_relations,
            shapes=self.interaction.relation_shape,
            label="relation",
        )
        # Comment: it is important that the regularizers are stored in a module list, in order to appear in
        # model.modules(). Thereby, we can collect them automatically.
        self.weight_regularizers = nn.ModuleList()
        # Explicitly call reset_parameters to trigger initialization
        self.reset_parameters_()

    def append_weight_regularizer(
        self,
        parameter: Union[str, nn.Parameter, Iterable[Union[str, nn.Parameter]]],
        regularizer: Regularizer,
    ) -> None:
        """Add a model weight to a regularizer's weight list, and register the regularizer with the model.

        :param parameter:
            The parameter, either as name, or as nn.Parameter object. A list of available parameter names is shown by
             `sorted(dict(self.named_parameters()).keys())`.
        :param regularizer:
            The regularizer instance which will regularize the weights.

        :raises KeyError: If an invalid parameter name was given
        """
        # normalize input
        if isinstance(parameter, (str, nn.Parameter)):
            parameter = [parameter]
        weights: Mapping[str, nn.Parameter] = dict(self.named_parameters())
        for param in parameter:
            if isinstance(param, str):
                if param not in weights:
                    raise KeyError(f"Invalid parameter_name={parameter}. Available are: {sorted(weights.keys())}.")
                param: nn.Parameter = weights[param]  # type: ignore
            regularizer.add_parameter(parameter=param)
        self.weight_regularizers.append(regularizer)

    def forward(
        self,
        h_indices: Optional[torch.LongTensor],
        r_indices: Optional[torch.LongTensor],
        t_indices: Optional[torch.LongTensor],
        slice_size: Optional[int] = None,
        slice_dim: Optional[str] = None,
    ) -> torch.FloatTensor:
        """Forward pass.

        This method takes head, relation and tail indices and calculates the corresponding score.

        All indices which are not None, have to be either 1-element or have the same shape, which is the batch size.

        :param h_indices:
            The head indices. None indicates to use all.
        :param r_indices:
            The relation indices. None indicates to use all.
        :param t_indices:
            The tail indices. None indicates to use all.
        :param slice_size:
            The slice size.
        :param slice_dim:
            The dimension along which to slice. From {"h", "r", "t"}

        :return: shape: (batch_size, num_heads, num_relations, num_tails)
            The score for each triple.
        """
        h, r, t = self._get_representations(h_indices=h_indices, r_indices=r_indices, t_indices=t_indices)
        scores = self.interaction.score(h=h, r=r, t=t, slice_size=slice_size, slice_dim=slice_dim)
        return self._repeat_scores_if_necessary(
            scores=scores, h_indices=h_indices, r_indices=r_indices, t_indices=t_indices,
        )

    def _repeat_scores_if_necessary(
        self,
        scores: torch.FloatTensor,
        h_indices: Optional[torch.LongTensor],
        r_indices: Optional[torch.LongTensor],
        t_indices: Optional[torch.LongTensor],
    ) -> torch.FloatTensor:
        """Repeat scores for entities/relations if the model does not have representations for one of them."""
        repeat_relations = len(self.relation_representations) == 0
        repeat_entities = len(self.entity_representations) == 0

        if not (repeat_entities or repeat_relations):
            return scores

        repeats = [1, 1, 1, 1]

        for i, (flag, ind, num) in enumerate((
            (repeat_entities, h_indices, self.num_entities),
            (repeat_relations, r_indices, self.num_relations),
            (repeat_entities, t_indices, self.num_entities),
        ), start=1):
            if flag:
                if ind is None:
                    repeats[i] = num
                else:
                    batch_size = len(ind)
                    if scores.shape[0] < batch_size:
                        repeats[0] = batch_size

        return scores.repeat(*repeats)

    def _get_representations(
        self,
        h_indices: Optional[torch.LongTensor],
        r_indices: Optional[torch.LongTensor],
        t_indices: Optional[torch.LongTensor],
    ) -> Tuple[HeadRepresentation, RelationRepresentation, TailRepresentation]:
        """Get representations for head, relation and tails, in canonical shape."""
        h, r, t = [
            [
                representation.get_in_more_canonical_shape(dim=dim, indices=indices)
                for representation in representations
            ]
            for dim, indices, representations in (
                ("h", h_indices, self.entity_representations),
                ("r", r_indices, self.relation_representations),
                ("t", t_indices, self.entity_representations),
            )
        ]
        # normalization
        return cast(
            Tuple[HeadRepresentation, RelationRepresentation, TailRepresentation],
            tuple(x[0] if len(x) == 1 else x for x in (h, r, t)),
        )



# -*- coding: utf-8 -*-

"""Tests for remixing and other triples reorganization."""

import unittest

import torch

from pykeen.triples import CoreTriplesFactory
from pykeen.triples.triples_factory import splits_similarity, splits_steps


class TestRemix(unittest.TestCase):
    """Tests for remix functions."""

    def test_splits_similarity(self):
        """Test the similarity calculation."""
        a_train = torch.as_tensor([
            [1, 1, 2],
            [2, 1, 3],
            [1, 2, 3],
            [4, 1, 5],
            [5, 1, 6],
        ])
        a_test = torch.as_tensor([
            [4, 2, 6],
        ])
        b_train = torch.as_tensor([
            [1, 1, 2],
            [2, 1, 3],
            [1, 2, 3],
            [4, 1, 5],
            [4, 2, 6],
        ])
        b_test = torch.as_tensor([
            [5, 1, 6],
        ])

        a_train_tf = CoreTriplesFactory.create(a_train)
        a_test_tf = CoreTriplesFactory.create(a_test)
        b_train_tf = CoreTriplesFactory.create(b_train)
        b_test_tf = CoreTriplesFactory.create(b_test)

        steps = splits_steps([a_train_tf, a_test_tf], [b_train_tf, b_test_tf])
        self.assertEqual(2, steps)

        similarity = splits_similarity([a_train_tf, a_test_tf], [b_train_tf, b_test_tf])
        self.assertEqual(1 - steps / 6, similarity)
